"""
Tests the DQM compute dispatch module
"""


import numpy as np
import pytest
from qcelemental.testing import compare_values

import qcengine as qcng
from qcengine.testing import checkver_and_convert, failure_engine, from_v2, schema_versions, uusing


def test_missing_key(schema_versions, request):
    _, retver, _ = schema_versions

    ret = qcng.compute({"hello": "hi"}, "bleh", return_version=retver)
    if "as_v2" not in request.node.name:
        # input provides no clue to v2/v1, so checkver fails
        ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.success is False
    assert "hello" in ret.input_data


def test_missing_key_raises(schema_versions, request):
    with pytest.raises(qcng.exceptions.InputError):
        ret = qcng.compute({"hello": "hi"}, "bleh", raise_error=True)


@uusing("psi4")
def test_psi4_task(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {"method": "SCF", "basis": "sto-3g"},
                "keywords": {"scf_type": "df"},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {"method": "SCF", "basis": "sto-3g"},
            "keywords": {"scf_type": "df"},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "psi4", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    if "v2" in request.node.name:
        assert ret.input_data.specification.driver == "energy"
    else:
        assert ret.driver == "energy"
    assert "Final Energy" in ret.stdout

    prov_keys = {"cpu", "hostname", "username", "wall_time"}
    assert ret.provenance.model_dump().keys() >= prov_keys
    assert "retries" not in ret.provenance.model_dump()

    assert ret.success is True


@uusing("psi4")
@uusing("classic-gcp")
def test_psi4_hf3c_task(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {"method": "HF3c"},
                "keywords": {"scf_type": "df"},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {"method": "HF3c"},
            "keywords": {"scf_type": "df"},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "psi4", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True
    if "v2" in request.node.name:
        # prior to 0.50, None, now ""
        assert not ret.input_data.specification.model.basis
    else:
        assert not ret.model.basis


@uusing("psi4")
def test_psi4_interactive_task(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {"method": "SCF", "basis": "sto-3g"},
                "keywords": {"scf_type": "df"},
                "extras": {"psiapi": True},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {"method": "SCF", "basis": "sto-3g"},
            "keywords": {"scf_type": "df"},
            "extras": {"psiapi": True},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "psi4", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert "Final Energy" in ret.stdout
    assert ret.success
    is_psiapi_evaluated = ret.extras.pop("psiapi_evaluated", False)
    assert is_psiapi_evaluated


@uusing("psi4")
def test_psi4_wavefunction_task(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {"method": "SCF", "basis": "sto-3g"},
                "protocols": {"wavefunction": "orbitals_and_eigenvalues"},
                "keywords": {"scf_type": "df"},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {"method": "SCF", "basis": "sto-3g"},
            "keywords": {"scf_type": "df"},
            "protocols": {"wavefunction": "orbitals_and_eigenvalues"},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "psi4", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success, ret.error.error_message
    assert ret.wavefunction.scf_orbitals_a.shape == (7, 7)


@uusing("psi4")
def test_psi4_internal_failure(schema_versions, request):
    models, retver, _ = schema_versions

    mol = models.Molecule.from_data(
        """0 3
     O    0.000000000000     0.000000000000    -0.068516245955
    """
    )

    if from_v2(request.node.name):
        psi4_task = {
            "molecule": mol,
            "specification": {
                "driver": "energy",
                "model": {"method": "ccsd", "basis": "6-31g"},
                "keywords": {"reference": "rhf"},
            },
        }
    else:
        psi4_task = {
            "molecule": mol,
            "driver": "energy",
            "model": {"method": "ccsd", "basis": "6-31g"},
            "keywords": {"reference": "rhf"},
        }
    with pytest.raises(qcng.exceptions.InputError) as exc:
        psi4_task = checkver_and_convert(psi4_task, request.node.name, "pre")
        qcng.compute(psi4_task, "psi4", raise_error=True, return_version=retver)

    assert "reference is only" in str(exc.value)


@uusing("psi4")
def test_psi4_ref_switch(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            **{
                "molecule": {"symbols": ["Li"], "geometry": [0, 0, 0], "molecular_multiplicity": 2},
                "specification": {
                    "driver": "energy",
                    "model": {"method": "B3LYP", "basis": "sto-3g"},
                    "keywords": {"scf_type": "df"},
                },
            }
        )
    else:
        inp = models.AtomicInput(
            **{
                "molecule": {"symbols": ["Li"], "geometry": [0, 0, 0], "molecular_multiplicity": 2},
                "driver": "energy",
                "model": {"method": "B3LYP", "basis": "sto-3g"},
                "keywords": {"scf_type": "df"},
            }
        )

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, "psi4", raise_error=True, return_dict=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True
    assert ret.properties.calcinfo_nalpha == 2
    assert ret.properties.calcinfo_nbeta == 1


@uusing("rdkit")
@pytest.mark.parametrize("method", ["UFF", "MMFF94", "MMFF94s"])
def test_rdkit_task(method, schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {"driver": "gradient", "model": {"method": method}, "keywords": {}},
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "gradient",
            "model": {"method": method},
            "keywords": {},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "rdkit", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True


@uusing("rdkit")
def test_rdkit_connectivity_error(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": qcng.get_molecule("water", return_dict=True),
            "specification": {"driver": "energy", "model": {"method": "UFF", "basis": ""}, "keywords": {}},
        }
    else:
        input_data = {
            "molecule": qcng.get_molecule("water", return_dict=True),
            "driver": "energy",
            "model": {"method": "UFF", "basis": ""},
            "keywords": {},
        }
    del input_data["molecule"]["connectivity"]

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "rdkit", return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.success is False
    assert "connectivity" in ret.error.error_message

    with pytest.raises(qcng.exceptions.InputError):
        qcng.compute(input_data, "rdkit", raise_error=True, return_version=retver)


@uusing("torchani")
def test_torchani_task(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {"driver": "gradient", "model": {"method": "ANI1x", "basis": None}, "keywords": {}},
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "gradient",
            "model": {"method": "ANI1x", "basis": None},
            "keywords": {},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "torchani", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True
    if "v2" in request.node.name:
        assert ret.input_data.specification.driver == "gradient"
    else:
        assert ret.driver == "gradient"


@uusing("mopac")
def test_mopac_task(schema_versions, request):
    _, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = {
            "molecule": qcng.get_molecule("water", return_dict=True),
            "specification": {
                "driver": "gradient",
                "model": {"method": "PM6", "basis": None},
                "keywords": {"pulay": False},
            },
        }
    else:
        input_data = {
            "molecule": qcng.get_molecule("water", return_dict=True),
            "driver": "gradient",
            "model": {"method": "PM6", "basis": None},
            "keywords": {"pulay": False},
        }

    input_data1 = checkver_and_convert(input_data.copy(), request.node.name, "pre")
    ret = qcng.compute(input_data1, "mopac", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.extras.keys() >= {"heat_of_formation", "dip_vec"}
    energy = pytest.approx(-0.08474117913025125, rel=1.0e-5)

    # Check gradient
    input_data2 = checkver_and_convert(input_data.copy(), request.node.name, "pre")
    ret = qcng.compute(input_data2, "mopac", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.extras.keys() >= {"heat_of_formation", "dip_vec"}
    assert np.linalg.norm(ret.return_result) == pytest.approx(0.03543560156912385, rel=3.0e-4)
    assert ret.properties.return_energy == energy

    # Check energy
    input_data3 = input_data.copy()
    if from_v2(request.node.name):
        input_data3["specification"]["driver"] = "energy"
    else:
        input_data3["driver"] = "energy"
    input_data3 = checkver_and_convert(input_data3, request.node.name, "pre")
    ret = qcng.compute(input_data3, "mopac", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.return_result == energy
    assert "== MOPAC DONE ==" in ret.stdout


def test_random_failure_no_retries_input(failure_engine, schema_versions, request):
    _, retver, _ = schema_versions

    failure_engine.iter_modes = ["input_error"]
    ret = qcng.compute(failure_engine.get_job(), failure_engine.name, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.error.error_type == "input_error"
    provenance_keys = (
        ret.input_data.provenance.model_dump().keys()
        if ("_v2" in request.node.name)
        else ret.input_data["provenance"].keys()
    )
    assert "retries" not in provenance_keys


def test_random_failure_no_retries_random(failure_engine, schema_versions, request):
    _, retver, _ = schema_versions

    failure_engine.iter_modes = ["random_error"]
    ret = qcng.compute(failure_engine.get_job(), failure_engine.name, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.error.error_type == "random_error"
    provenance_keys = (
        ret.input_data.provenance.model_dump().keys()
        if ("_v2" in request.node.name)
        else ret.input_data["provenance"].keys()
    )
    assert "retries" not in provenance_keys


def test_random_failure_with_retries(failure_engine, schema_versions, request):
    _, retver, _ = schema_versions

    failure_engine.iter_modes = ["random_error", "random_error", "random_error"]
    ret = qcng.compute(
        failure_engine.get_job(),
        failure_engine.name,
        raise_error=False,
        task_config={"retries": 2},
        return_version=retver,
    )
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    retries = (
        ret.input_data.provenance.retries if ("_v2" in request.node.name) else ret.input_data["provenance"]["retries"]
    )
    assert retries == 2
    assert ret.error.error_type == "random_error"

    failure_engine.iter_modes = ["random_error", "input_error"]
    ret = qcng.compute(
        failure_engine.get_job(),
        failure_engine.name,
        raise_error=False,
        task_config={"retries": 4},
        return_version=retver,
    )
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    retries = (
        ret.input_data.provenance.retries if ("_v2" in request.node.name) else ret.input_data["provenance"]["retries"]
    )
    assert retries == 1
    assert ret.error.error_type == "input_error"


def test_random_failure_with_success(failure_engine, schema_versions, request):
    _, retver, _ = schema_versions

    failure_engine.iter_modes = ["random_error", "pass"]
    failure_engine.ncalls = 0
    ret = qcng.compute(
        failure_engine.get_job(),
        failure_engine.name,
        raise_error=False,
        task_config={"retries": 1},
        return_version=retver,
    )
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success, ret.error.error_message
    assert ret.provenance.retries == 1
    assert ret.extras["ncalls"] == 2


@uusing("openmm")
def test_openmm_task_smirnoff(schema_versions, request):
    models, retver, _ = schema_versions

    from qcengine.programs.openmm import OpenMMHarness

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {"method": "openff-1.0.0", "basis": "smirnoff"},
                "keywords": {},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {"method": "openff-1.0.0", "basis": "smirnoff"},
            "keywords": {},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "openmm", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    cachelength = len(OpenMMHarness._CACHE)

    assert cachelength > 0
    assert ret.success is True

    ret = qcng.compute(input_data, "openmm", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    # ensure cache has not grown
    assert len(OpenMMHarness._CACHE) == cachelength
    assert ret.success is True


@pytest.mark.skip("`basis` must be explicitly specified at this time")
@uusing("openmm")
def test_openmm_task_url_basis(schema_versions, request):
    models, retver, _ = schema_versions

    from qcengine.programs.openmm import OpenMMHarness

    if from_v2(request.node.name):
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "specification": {
                "driver": "energy",
                "model": {
                    "method": "openmm",
                    "basis": "openff-1.0.0",
                    "url": "https://raw.githubusercontent.com/openforcefield/openff-forcefields/1.0.0/openforcefields/offxml/openff-1.0.0.offxml",
                },
                "keywords": {},
            },
        }
    else:
        input_data = {
            "molecule": models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            "driver": "energy",
            "model": {
                "method": "openmm",
                "basis": "openff-1.0.0",
                "url": "https://raw.githubusercontent.com/openforcefield/openff-forcefields/1.0.0/openforcefields/offxml/openff-1.0.0.offxml",
            },
            "keywords": {},
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "openmm", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    cachelength = len(OpenMMHarness._CACHE)

    assert cachelength > 0
    assert ret.success is True

    ret = qcng.compute(input_data, "openmm", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    # ensure cache has not grown
    assert len(OpenMMHarness._CACHE) == cachelength
    assert ret.success is True


@uusing("openmm")
def test_openmm_cmiles_gradient(schema_versions, request):
    models, retver, _ = schema_versions

    program = "openmm"

    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))

    water_dict = water.model_dump()
    # add water cmiles to the molecule
    water_dict["extras"] = {"cmiles": {"canonical_isomeric_explicit_hydrogen_mapped_smiles": "[H:2][O:1][H:3]"}}

    molecule = models.Molecule.from_data(water_dict)

    model = {"method": "openff-1.0.0", "basis": "smirnoff"}

    if from_v2(request.node.name):
        inp = models.AtomicInput(molecule=molecule, specification={"driver": "gradient", "model": model})
    else:
        inp = models.AtomicInput(molecule=molecule, driver="gradient", model=model)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True


@uusing("openmm")
def test_openmm_cmiles_gradient_nomatch(schema_versions, request):
    models, retver, _ = schema_versions

    program = "openmm"

    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))

    water_dict = water.model_dump()
    # add ethane cmiles to the molecule
    water_dict["extras"] = {
        "cmiles": {
            "canonical_isomeric_explicit_hydrogen_mapped_smiles": "[H:3][C:1]([H:4])([H:5])[C:2]([H:6])([H:7])[H:8]"
        }
    }

    molecule = models.Molecule.from_data(water_dict)

    model = {"method": "openff-1.0.0", "basis": "smirnoff"}

    if from_v2(request.node.name):
        inp = models.AtomicInput(molecule=molecule, specification={"driver": "gradient", "model": model})
    else:
        inp = models.AtomicInput(molecule=molecule, driver="gradient", model=model)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    # if we correctly find the cmiles this should fail as the molecule and cmiles are different
    assert ret.success is False
    assert (
        "molecule.add_conformer given input of the wrong shape: Given (3, 3), expected (8, 3)"
        in ret.error.error_message
    )


@uusing("openmm")
@pytest.mark.parametrize(
    "gaff_settings",
    [
        pytest.param(({}, None, 0.0013904199062156914), id="gaff no keywords"),
        pytest.param(({"constraints": "ALLBONDS"}, None, 8.108238580315493e-05), id="constraints allbonds"),
        pytest.param(({"nonbondedMethod": "LjPmE"}, None, 0.0013904199062156914), id="nonbonded ljpme"),
        pytest.param(
            ({"nonbondedMethod": "PME", "constraints": "Hbonds"}, None, 8.108238580315493e-05),
            id="nonbonded pme constraints hbonds",
        ),
        pytest.param(({"constraints": "badmethod"}, ValueError, 0), id="incorrect constraint"),
        pytest.param(({"nonbondedMethod": "badmethod"}, ValueError, 0), id="incorrect nonbondedmethod"),
    ],
)
def test_openmm_gaff_keywords(gaff_settings, schema_versions, request):
    """
    Test the different running settings with gaff.
    """
    models, retver, _ = schema_versions

    program = "openmm"
    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))

    water_dict = water.model_dump()
    # add water cmiles to the molecule
    water_dict["extras"] = {"cmiles": {"canonical_isomeric_explicit_hydrogen_mapped_smiles": "[H:2][O:1][H:3]"}}

    molecule = models.Molecule.from_data(water_dict)
    keywords, error, expected_result = gaff_settings
    model = {"method": "gaff-2.1", "basis": "antechamber"}
    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule, specification={"driver": "energy", "model": model, "keywords": keywords}
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)
    if error is not None:
        with pytest.raises(error):
            inp = checkver_and_convert(inp, request.node.name, "pre")
            _ = qcng.compute(inp, program, raise_error=True, return_version=retver)
    else:
        inp = checkver_and_convert(inp, request.node.name, "pre")
        ret = qcng.compute(inp, program, raise_error=False, return_version=retver)
        ret = checkver_and_convert(ret, request.node.name, "post")

        assert ret.success is True
        assert ret.return_result == pytest.approx(expected_result, rel=1e-6)


@uusing("mace")
def test_mace_energy(schema_versions, request):
    """
    Test calculating the energy with mace
    """
    models, retver, _ = schema_versions
    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=water, specification={"driver": "energy", "model": {"method": "small", "basis": None}}
        )
    else:
        atomic_input = models.AtomicInput(molecule=water, model={"method": "small", "basis": None}, driver="energy")

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    result = qcng.compute(atomic_input, "mace", return_version=retver)
    result = checkver_and_convert(result, request.node.name, "post")

    assert result.success
    assert pytest.approx(result.return_result) == -76.47683956098838


@uusing("mace")
def test_mace_gradient(schema_versions, request):
    """
    Test calculating the gradient with mace
    """
    models, retver, _ = schema_versions

    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    expected_result = np.array(
        [
            [0.0, -2.1590400539385646e-18, -0.04178551770271103],
            [0.0, -0.029712483642769006, 0.020892758851355515],
            [0.0, 0.029712483642769006, 0.020892758851355518],
        ]
    )

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=water, specification={"driver": "gradient", "model": {"method": "small", "basis": None}}
        )
    else:
        atomic_input = models.AtomicInput(molecule=water, model={"method": "small", "basis": None}, driver="gradient")

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    result = qcng.compute(atomic_input, "mace", return_version=retver)
    result = checkver_and_convert(result, request.node.name, "post")

    assert result.success
    assert pytest.approx(result.return_result) == expected_result


@uusing("aimnet2")
@pytest.mark.parametrize(
    "model, expected_energy",
    [
        pytest.param("b973c", -76.39604306960972, id="b973c"),
        pytest.param("wb97m-d3", -76.47412023758551, id="wb97m-d3"),
    ],
)
def test_aimnet2_energy(model, expected_energy, schema_versions, request):
    """Test computing the energies of water with two aimnet2 models."""
    models, retver, _ = schema_versions

    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=water, specification={"driver": "energy", "model": {"method": model, "basis": None}}
        )
    else:
        atomic_input = models.AtomicInput(molecule=water, model={"method": model, "basis": None}, driver="energy")

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    result = qcng.compute(atomic_input, "aimnet2", return_version=retver)
    result = checkver_and_convert(result, request.node.name, "post")

    assert result.success
    assert pytest.approx(result.return_result) == expected_energy
    assert "charges" in result.extras["aimnet2"]
    assert "ensemble_charges_std" in result.extras["aimnet2"]
    assert "ensemble_forces_std" in result.extras["aimnet2"]


@uusing("aimnet2")
def test_aimnet2_gradient(schema_versions, request):
    """Test computing the gradient of water using one aimnet2 model."""
    models, retver, _ = schema_versions

    water = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=water, specification={"driver": "gradient", "model": {"method": "wb97m-d3", "basis": None}}
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=water, model={"method": "wb97m-d3", "basis": None}, driver="gradient"
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    result = qcng.compute(atomic_input, "aimnet2", return_version=retver)
    result = checkver_and_convert(result, request.node.name, "post")

    assert result.success
    # make sure the gradient is now the return result
    assert np.allclose(
        result.return_result,
        np.array(
            [
                [-0.0, 2.6080331227973375e-09, -0.04097248986363411],
                [-0.0, -0.029529934749007225, 0.020486244931817055],
                [-0.0, 0.029529931023716927, 0.020486244931817055],
            ]
        ),
    )
    assert pytest.approx(result.properties.return_energy) == -76.47412023758551
    # make sure the other properties were also saved
    assert "charges" in result.extras["aimnet2"]


@uusing("psi4")
def test_psi4_properties_driver(schema_versions, request):
    models, retver, _ = schema_versions

    import numpy as np

    json_data = {
        "schema_name": "qcschema_input",
        "schema_version": 1,
        "molecule": {
            "geometry": [
                0.0,
                0.0,
                -0.1294769411935893,
                0.0,
                -1.494187339479985,
                1.0274465079245698,
                0.0,
                1.494187339479985,
                1.0274465079245698,
            ],
            "symbols": ["O", "H", "H"],
            "fix_com": True,
            "fix_orientation": True,
        },
        "driver": "properties",
        "model": {
            "method": "HF",
            "basis": "6-31G",
            "properties": [
                "dipole",
                "quadrupole",
                "mulliken_charges",
                "lowdin_charges",
                "wiberg_lowdin_indices",
                "mayer_indices",
            ],
        },
        "keywords": {"scf_type": "df", "mp2_type": "df", "e_convergence": 9},
    }
    if from_v2(request.node.name):
        json_data["schema_name"] = "qcschema_atomic_input"
        json_data["schema_version"] = 2
        json_data["specification"] = {
            "driver": json_data.pop("driver"),
            "model": json_data.pop("model"),
            "keywords": json_data.pop("keywords"),
        }

    # Write expected output (dipole & quadrupole in au)
    expected_return_result = {
        "dipole": np.array([0.0, 0.0, 1.04037263345]).reshape((3,)),
        "quadrupole": np.array([-5.42788218076, 0.0, 0.0, 0.0, -3.07521129293, 0.0, 0.0, 0.0, -4.36605314966]).reshape(
            (3, 3)
        ),
        "mulliken_charges": [-0.7967275695997689, 0.3983637847998823, 0.3983637847998822],
        "lowdin_charges": [-0.5945105406840803, 0.29725527034203636, 0.29725527034203636],
        "wiberg_lowdin_indices": np.array(
            [
                0.0,
                0.9237385044125344,
                0.9237385044125329,
                0.9237385044125344,
                0.0,
                0.006992650019441531,
                0.9237385044125329,
                0.006992650019441531,
                0.0,
            ]
        ).reshape((3, 3)),
        "mayer_indices": np.array(
            [
                0.0,
                0.802064044935596,
                0.8020640449355959,
                0.802064044935596,
                0.0,
                0.003020025778524219,
                0.8020640449355959,
                0.003020025778524219,
                0.0,
            ]
        ).reshape((3, 3)),
    }

    expected_properties = {
        "calcinfo_nbasis": 13,
        "calcinfo_nmo": 13,
        "calcinfo_nalpha": 5,
        "calcinfo_nbeta": 5,
        "calcinfo_natom": 3,
        "scf_one_electron_energy": -122.27509111304202,
        "scf_two_electron_energy": 37.49348718008625,
        "nuclear_repulsion_energy": 8.80146206062943,
        "scf_total_energy": -75.98014187232634,
        "return_energy": -75.98014187232634,
    }

    json_data = checkver_and_convert(json_data, request.node.name, "pre")
    json_ret = qcng.compute(json_data, "psi4", return_version=retver)
    json_ret = checkver_and_convert(json_ret, request.node.name, "post")

    assert json_ret.success
    xptd_schema_name = "qcschema_atomic_result" if "v2" in request.node.name else "qcschema_output"
    assert json_ret.schema_name == xptd_schema_name
    for k in expected_return_result.keys():
        assert compare_values(expected_return_result[k], json_ret.return_result[k], atol=1.0e-5)
    for k in expected_properties.keys():
        assert compare_values(expected_properties[k], getattr(json_ret.properties, k), atol=1.0e-5)



"""
Integrates the computes together
"""
import os
import sys
import warnings
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import qcelemental

from .config import get_config
from .exceptions import InputError, RandomError
from .procedures import get_procedure
from .programs import get_program
from .util import QCEL_V1V2_SHIM_CODE, compute_wrapper, environ_context, handle_output_metadata

if TYPE_CHECKING:
    from pydantic.main import BaseModel
    from qcelemental.models.v2 import FailedOperation


__all__ = ["compute", "compute_procedure"]


def _process_failure_and_return(model, return_dict, raise_error):
    if isinstance(model, (qcelemental.models.v1.FailedOperation, qcelemental.models.v2.FailedOperation)):
        if raise_error:
            raise InputError(model.error.error_message)
        elif return_dict:
            return model.model_dump()
        else:
            return model
    else:
        return False


def compute(
    input_data: Union[Dict[str, Any], "BaseModel"],  # TODO Input base class
    program: str,
    raise_error: bool = False,
    task_config: Optional[Dict[str, Any]] = None,
    return_dict: bool = False,
    return_version: int = -1,
) -> Union[
    "BaseModel", "FailedOperation", Dict[str, Any]
]:  # TODO Output base class, was AtomicResult OptimizationResult
    """Executes a single CMS program given a QCSchema input.

    The full specification can be found at:
        http://molssi-qc-schema.readthedocs.io/en/latest/index.html#

    Parameters
    ----------
    input_data
        A QCSchema input specification in dictionary or model from QCElemental.models
    program
        The CMS program or procedure with which to execute the input. E.g., "psi4", "rdkit", "geometric".
    raise_error
        If computation doesn't succeed, should compute raise an error (`True`) or encode in a
        `FailedOperation` (`False`). See also: `return_dict`.
    retries : int, optional
        The number of random tries to retry for.
    task_config
        A dictionary of local configuration options corresponding to a TaskConfig object.
        Formerly local_options.
    return_dict
        Returns a dictionary serialization instead of a Result or FailedOperation model from
        QCElemental.models. Note that `raise_error` take precedence.
    return_version
        The schema version to return. If -1, the input schema_version is used.

    Returns
    -------
    result
        AtomicResult, OptimizationResult, FailedOperation, etc., or Dict representation of any object type
        A QCSchema representation of the requested output, type depends on return_dict key.

        .. _`table:compute_result`:

        +-----------+-------------+-------------+--------------------------------------------------+
        | good_calc | raise_error | return_dict | output                                           |
        +===========+=============+=============+==================================================+
        | T         | F (def)     | F (def)     | ``AtomicResult`` object                          |
        +-----------+-------------+-------------+--------------------------------------------------+
        | T         | T           | F (def)     | ``AtomicResult`` object                          |
        +-----------+-------------+-------------+--------------------------------------------------+
        | T         | T           | T           | dict of ``AtomicResult``                         |
        +-----------+-------------+-------------+--------------------------------------------------+
        | T         | F (def)     | T           | dict of ``AtomicResult``                         |
        +-----------+-------------+-------------+--------------------------------------------------+
        | F         | F (def)     | F (def)     | ``FailedOperation`` object                       |
        +-----------+-------------+-------------+--------------------------------------------------+
        | F         | T           | F (def)     | raises ``InputError`` (type encoded in FailedOp) |
        +-----------+-------------+-------------+--------------------------------------------------+
        | F         | T           | T           | raises ``InputError`` (type encoded in FailedOp) |
        +-----------+-------------+-------------+--------------------------------------------------+
        | F         | F (def)     | T           | dict of ``FailedOperation``                      |
        +-----------+-------------+-------------+--------------------------------------------------+

        .. _`table:compute_result_schver`:

        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | good_calc? | input_data ver | return_version | return_dict | output                                                                 |
        +============+================+================+=============+========================================================================+
        | T          | 1              | -1 (def) or 1  | F (def)     | ``v1.AtomicResult`` object (not avail. Py 3.14+ ^^)                    |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 1              |  2             | F (def)     | ``v2.AtomicResult`` object                                             |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 1              | -1 (def) or 1  | T           | dict of ``v1.AtomicResult`` @@                                         |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 1              |  2             | T           | dict of ``v2.AtomicResult``                                            |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 2              | -1 (def) or 2  | F (def)     | ``v2.AtomicResult`` object                                             |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 2              |  1             | F (def)     | ``v1.AtomicResult`` object (not avail. Py 3.14+ ^^)                    |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 2              | -1 (def) or 2  | T           | dict of ``v2.AtomicResult``                                            |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | T          | 2              |  1             | T           | dict of ``v1.AtomicResult`` @@                                         |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 1              | -1 (def) or 1  | F (def)     | ``v1.FailedOperation`` object (not avail. Py 3.14+ ^^) **              |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 1              |  2             | F (def)     | ``v2.FailedOperation`` object                                          |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 1              | -1 (def) or 1  | T           | dict of ``v1.FailedOperation`` @@ **                                   |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 1              |  2             | T           | dict of ``v2.FailedOperation``                                         |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 2              | -1 (def) or 2  | F (def)     | ``v2.FailedOperation`` object **                                       |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 2              |  1             | F (def)     | ``v1.FailedOperation`` object (not avail. Py 3.14+ ^^)                 |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 2              | -1 (def) or 2  | T           | dict of ``v2.FailedOperation`` **                                      |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | F          | 2              |  1             | T           | dict of ``v1.FailedOperation`` @@                                      |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        | ** F       |                | ** -1          |             | for errors *before* input ver detected, returns v2 if Py 3.14+ else v1 |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        |            |                | ^^ 1           | ^^ F        | for Py 3.14+, returns v2.FailedOp for minimal API surprise             |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+
        |            |                | @@ 1           | @@ T        | for Py 3.14+, returns dict of v1 (like all other Py) with brittle      |
        +            +                +                +             +                                                                        +
        |            |                |                |             |   :envvar:`QCNG_USE_V1V2_SHIM=1`, otherwise dict of v2.FailedOp        |
        +------------+----------------+----------------+-------------+------------------------------------------------------------------------+

    .. versionadded:: 0.50.0
       input_data can newly be QCSchema v2 as well as longstanding v1.
       Note that the QCSchema v2 layout will not be official until v0.60.0 .
       The compute_procedure is newly incorporated into compute.
       The *return_version* parameter was added.

    """

    try:
        # models, v1 or v2
        output_data = input_data.model_copy()
    except AttributeError:
        # dicts
        output_data = input_data.copy()  # lgtm [py/multiple-definition]

    with compute_wrapper(capture_output=False, raise_error=raise_error) as metadata:
        # Grab the executor harness
        try:
            executor = get_procedure(program)
        except InputError:
            executor = get_program(program)

        # Build the model and validate
        # * calls model_wrapper with the (Atomic|Optimization|etc)Input for which the harness was designed
        # * upon return, input_data is a model of the type (e.g., Atomic) and version (e.g., 1 or 2) the harness prefers: all v2.
        input_data, input_schema_version = executor.build_input_model(input_data, return_input_schema_version=True)
        return_version = input_schema_version if return_version == -1 else return_version

        # V1V2TEST if return_version == 1:
        if sys.version_info >= (3, 14) and return_version == 1:  # forbidden by pydantic ...

            _MSG314 = (
                f"QCSchema v1 models (this: {input_data.schema_name}) cannot be instantiated in this environment. "
                + "Reason: pydantic.v1 is unavailable on Python 3.14+. You can: "
                + "(a) use Python <3.14; or "
                + "(b) use QCSchema v2 (either input or ask for output with `return_version=2`; "
                + " note that QCSchema v2 not finalized until QCElemental v0.60); or "
                + "(c) ask for a QCSchema v1 dictionary back rather than the model with `return_dict=True` "
                + " while setting envvar `QCNG_USE_V1V2_SHIM=1` to acknowledge this is brittle."
            )

            # V1V2TEST if input_data.schema_name == "qcschema_atomic_input":
            if (
                return_dict is True and input_data.schema_name == "qcschema_atomic_input"
            ):  # ... but we have a workaround ...

                if bool(os.environ.get("QCNG_USE_V1V2_SHIM", False)):  # ... if choose to use it
                    # return_version = -12 = QCEL_V1V2_SHIM_CODE signals to use the shim classes that represent certain
                    #   QCSchema v1 layouts (QCSk v1 only exist in pydantic.v1 API) in pydantic v2 API.
                    #   We never want to release these into the wild so only available if returning
                    #   dict and only for Atomic models (i.e., shims not avail for TD, MBE; shim avail for Opt but not enabled here).
                    return_version = QCEL_V1V2_SHIM_CODE
                else:
                    # reset retver so that FailedOp w/_MSG314 can be constructed
                    return_version = 2
                    raise RuntimeError(_MSG314)
            else:
                # reset retver so that FailedOp w/_MSG314 can be constructed
                return_version = 2
                raise RuntimeError(_MSG314)

        # Build out task_config
        if task_config is None:
            task_config = {}
        input_engine_options = input_data.specification.extras.pop("_qcengine_local_config", {})
        task_config = {**task_config, **input_engine_options}
        config = get_config(task_config=task_config)

        # Set environment parameters and execute
        with environ_context(config=config):

            # Handle optional retries
            for x in range(config.retries + 1):
                try:
                    output_data = executor.compute(input_data, config)
                    break
                except RandomError as e:
                    if return_version >= 2:
                        output_data = input_data

                    if x == config.retries:
                        raise e
                    else:
                        metadata["retries"] += 1
                except Exception:
                    if return_version >= 2:
                        output_data = input_data
                    raise

    return handle_output_metadata(
        output_data, metadata, raise_error=raise_error, return_dict=return_dict, convert_version=return_version
    )


def compute_procedure(*args, **kwargs):
    from qcelemental.models.common_models import _qcsk_v2_default_v1_importpathschange

    warnings.warn(
        "Using the `compute_procedure` function is deprecated in favor of using `compute`, "
        f"and as soon as version {_qcsk_v2_default_v1_importpathschange} it will stop working.",
        category=FutureWarning,
        stacklevel=2,
    )
    if "procedure" in kwargs:
        kwargs["program"] = kwargs.pop("procedure")
    return compute(*args, **kwargs)



"""
Tests the DQM compute dispatch module
"""

import pytest

import qcengine as qcng
from qcengine.testing import using


def test_list_programs():

    r = qcng.list_all_programs()
    assert r >= {"psi4", "rdkit", "molpro", "dftd3"}


@pytest.mark.parametrize(
    "program",
    [
        pytest.param("psi4", marks=using("psi4")),
        pytest.param("torchani", marks=using("torchani")),
        pytest.param("rdkit", marks=using("rdkit")),
    ],
)
def test_check_program_avail(program):

    assert program in qcng.list_available_programs()


def test_program_version_check_is_targeted(monkeypatch):
    """Ensure version checks probe only the requested program harness."""
    from qcengine.testing import is_program_new_enough

    class AvailableHarness:
        @staticmethod
        def found():
            return True

        @staticmethod
        def get_version():
            return "2.0"

    def unexpected_full_scan():
        pytest.fail("is_program_new_enough should not scan every registered program")

    monkeypatch.setattr(qcng, "list_all_procedures", set)
    monkeypatch.setattr(qcng, "list_available_programs", unexpected_full_scan)
    monkeypatch.setattr(qcng, "get_program", lambda name, check=False: AvailableHarness())

    assert is_program_new_enough("available", "1.0")


def test_program_avail_bounce():

    with pytest.raises(qcng.exceptions.InputError) as exc:
        qcng.compute({}, "bad_program", raise_error=True)

    assert "not registered" in str(exc.value)


def test_list_procedures():

    r = qcng.list_all_procedures()
    assert r >= {"geometric", "berny"}


@pytest.mark.parametrize(
    "procedure", [pytest.param("geometric", marks=using("geometric")), pytest.param("berny", marks=using("berny"))]
)
def test_check_procedure_avail(procedure):

    assert procedure in qcng.list_available_procedures()


def test_procedure_avail_bounce():

    with pytest.raises(qcng.exceptions.InputError) as exc:
        qcng.compute({}, "bad_program", raise_error=True)

    assert "not registered" in str(exc.value)



"""
Calls the TorchANI package.
"""

from typing import TYPE_CHECKING, Any, ClassVar, Dict

from qcelemental.models.v2 import AtomicResult, Provenance
from qcelemental.util import parse_version, safe_version, which_import

from ..exceptions import InputError, ResourceError
from ..units import ureg
from .model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models.v2 import AtomicInput

    from ..config import TaskConfig


class TorchANIHarness(ProgramHarness):
    """Interface for TorchANI project."""

    _CACHE = {}

    _defaults: ClassVar[Dict[str, Any]] = {
        "name": "TorchANI",
        "scratch": False,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which_import(
            "torchani",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `pip install torchani`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("torchani")
        if which_prog not in self.version_cache:
            import torchani

            self.version_cache[which_prog] = safe_version(torchani.__version__)

        return self.version_cache[which_prog]

    def get_model(self, name: str) -> "torchani.models.BuiltinModels":
        name = name.lower()

        if name in self._CACHE:
            return self._CACHE[name]

        import torch
        import torch.nn as nn
        import torchani

        class EnsembleEnergies(nn.Module):
            """
            Wrap a TorchANI model and return (species, concatenated_member_energies).

            Compatible with both call styles:
              - m((species, coordinates))
              - m(species, coordinates, *extra_args)
            """

            def __init__(self, base_model):
                super().__init__()
                self.base = base_model

                # expose `model.species` for v2.2 for symbol->index conversion
                if hasattr(base_model, "species"):
                    self.species = base_model.species

            @staticmethod
            def _normalize_call_args(first, rest):
                # forward((species, coords), ...)
                if isinstance(first, (tuple, list)) and len(first) == 2:
                    species, coords = first
                    extra = rest
                else:
                    # forward(species, coords, ...)
                    species = first
                    coords = rest[0]
                    extra = rest[1:]
                return species, coords, extra

            @staticmethod
            def _extract_energies(member_result):
                # TorchANI 2.7+: returns object with .energies
                if hasattr(member_result, "energies"):
                    return member_result.energies
                # TorchANI 2.2.x (and older style): returns (species, energies)
                if isinstance(member_result, (tuple, list)) and len(member_result) >= 2:
                    return member_result[1]
                raise TypeError(f"Unrecognized TorchANI model output: {type(member_result)!r}")

            def forward(self, first, *rest, **kwargs):
                species, coords, extra = self._normalize_call_args(first, rest)

                member_energies = []
                n_members = len(self.base)
                for i in range(n_members):
                    member = self.base[i]
                    out = member((species, coords), *extra, **kwargs)
                    e = self._extract_energies(out)
                    member_energies.append(e)

                # keep old flat vector pattern using torch.cat([...])
                outputs = torch.cat(member_energies, dim=0)
                return species, outputs

        ani_models = {
            "ani1x": torchani.models.ANI1x,
            "ani1ccx": torchani.models.ANI1ccx,
        }
        if parse_version(self.get_version()) >= parse_version("2.0"):
            ani_models["ani2x"] = torchani.models.ANI2x

        if name not in ani_models:
            raise InputError(f"TorchANI only accepts methods: {list(ani_models.keys())}")

        base = ani_models[name]()  # the actual TorchANI model
        wrapped = EnsembleEnergies(base)  # your compatibility wrapper
        self._CACHE[name] = wrapped

        return self._CACHE[name]

    def compute(self, input_data: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        """
        Runs TorchANI in FF typing
        """
        # Check if exists and version
        self.found(raise_error=True)
        if parse_version(self.get_version()) < parse_version("0.9"):
            raise ResourceError("QCEngine's TorchANI wrapper requires version 0.9 or greater.")

        import numpy as np
        import torch
        import torchani

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Failure flag
        ret_data = {"success": False}

        # helpers functions for version compatibility
        def species_tensor_from_symbols(symbols, model, device):
            """
            Returns a (1, A) LongTensor species tensor compatible with both:
              - torchani 2.7+ (atomic numbers input)
              - torchani 2.2.x (model-index input via model.species)
            """
            symbols = list(symbols)

            # Newer TorchANI (e.g., 2.7+) supports atomic-number conversion in torchani.utils
            to_Z = getattr(torchani.utils, "ChemicalSymbolsToAtomicNumbers", None)
            if to_Z is not None:
                return to_Z()(symbols).to(device).unsqueeze(0)

            # Older TorchANI (e.g., 2.2.4): convert to model indices using model.species
            from torchani.utils import ChemicalSymbolsToInts

            if not hasattr(model, "species"):
                raise RuntimeError(
                    "TorchANI version lacks ChemicalSymbolsToAtomicNumbers, and model has no .species; "
                    "cannot build species tensor."
                )
            return ChemicalSymbolsToInts(model.species)(symbols).to(device).unsqueeze(0)

        def get_ensemble_energies(model_result):
            """
            Return a 1D tensor of ensemble member energies for a single geometry.
            Works with:
              - TorchANI result objects: result.energies
              - old style tuples: (species, energies)
              - your current wrapper tuple output
            """
            if hasattr(model_result, "energies"):
                return model_result.energies
            if isinstance(model_result, (tuple, list)) and len(model_result) >= 2:
                return model_result[1]
            raise TypeError(f"Unrecognized TorchANI model output: {type(model_result)!r}")

        # Build model
        method = input_data.specification.model.method
        model = self.get_model(method).to(device)

        # Validate and build species tensor
        symbols = list(input_data.molecule.symbols)

        known_sym = {"H", "C", "N", "O"}
        if method.lower() == "ani2x":
            known_sym.update({"S", "F", "Cl"})
        unknown_sym = set(symbols) - known_sym
        if unknown_sym:
            raise InputError(f"TorchANI model '{method}' does not support symbols: {unknown_sym}.")

        num_atoms = len(symbols)
        species = species_tensor_from_symbols(symbols, model, device)

        # Build coord array -- single geometry, batch dim = 1
        geom_array = input_data.molecule.geometry.reshape(1, -1, 3) * ureg.conversion_factor("bohr", "angstrom")

        # Keep gradients enabled for gradient/hessian drivers
        coordinates = torch.tensor(geom_array.tolist(), device=device, requires_grad=True)

        # Run model
        result = model((species, coordinates))
        ensemble_energies = get_ensemble_energies(result)

        # Ensure 1D member energies for "no batching" usage.
        # Some paths might return shape (n_members, 1) or (n_members,) — normalize to (n_members,)
        if ensemble_energies.ndim == 2 and ensemble_energies.shape[-1] == 1:
            ensemble_energies = ensemble_energies[:, 0]

        # Scalar energy used for derivatives: mean across ensemble members
        energy = ensemble_energies.mean()

        ensemble_std = ensemble_energies.std(unbiased=False)
        ensemble_scaled_std = ensemble_std / np.sqrt(num_atoms)

        ret_data["properties"] = {"return_energy": energy.item()}

        if input_data.specification.driver == "energy":
            ret_data["return_result"] = ret_data["properties"]["return_energy"]
        elif input_data.specification.driver == "gradient":
            derivative = torch.autograd.grad(energy, coordinates, create_graph=False, retain_graph=False)[0].squeeze()
            ret_data["return_result"] = (
                np.asarray(derivative.cpu() * ureg.conversion_factor("angstrom", "bohr")).ravel().tolist()
            )
        elif input_data.specification.driver == "hessian":
            # torchani.utils.hessian expects coordinates and energies (scalar or (batch,))
            hessian = torchani.utils.hessian(coordinates, energies=energy)
            ret_data["return_result"] = np.asarray(hessian.cpu())
        else:
            raise InputError(
                f"TorchANI can only compute energy, gradient, and hessian driver methods. Found {input_data.specification.driver}."
            )

        #######################################################################
        # Description of the quantities stored in `extras`
        #
        # ensemble_energies:
        #   An energy array of all members (models) in an ensemble of models
        #
        # ensemble_energy_avg:
        #   The average value of energy array which is also recorded with as
        #   `energy` in QCEngine
        #
        # ensemble_energy_std:
        #   The standard deviation of energy array
        #
        # ensemble_per_root_atom_disagreement:
        #   The standard deviation scaled by the square root of N, with N being
        #   the number of atoms in the molecule. This is the quantity used in
        #   the query-by-committee (QBC) process in active learning to infer
        #   the reliability of the models in an ensemble, and produce more data
        #   points in the regions where this quantity is below a certain
        #   threshold (inclusion criteria)
        ret_data["input_data"] = input_data
        ret_data["molecule"] = input_data.molecule
        ret_data["extras"] = {
            # keep 1D array of ensemble member energies (no batching)
            "ensemble_energies": ensemble_energies.detach().cpu().numpy(),
            "ensemble_energy_avg": energy.item(),
            "ensemble_energy_std": ensemble_std.item(),
            "ensemble_per_root_atom_disagreement": ensemble_scaled_std.item(),
        }

        ret_data["provenance"] = Provenance(
            creator="torchani", version="unknown", routine="torchani.builtin.aev_computer"
        )

        ret_data["schema_name"] = "qcschema_atomic_result"
        ret_data["success"] = True

        return AtomicResult(**ret_data)



import pprint

import pytest
from qcelemental import constants
from qcelemental.testing import compare, compare_values

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2
from qcengine.testing import schema_versions as schema_versionsALL5
from qcengine.testing import schema_versions0, using, uusing

try:
    import qcmanybody as qcmb
except ModuleNotFoundError:
    schema_versions = schema_versions0
except RuntimeError:
    # py314 + qcmanybody 0.5.1
    schema_versions = schema_versions0
else:
    if qcmb.__version__.startswith("0.5."):
        schema_versions = schema_versions0
    else:
        schema_versions = schema_versionsALL5


@pytest.fixture
def he_tetramer(schema_versions, request):
    models, _, _ = schema_versions

    a2 = 2 / constants.bohr2angstroms
    return models.Molecule(
        symbols=["He", "He", "He", "He"],
        fragments=[[0], [1], [2], [3]],
        geometry=[0, 0, 0, 0, 0, a2, 0, a2, 0, 0, a2, a2],
    )


@uusing("qcmanybody")
@pytest.mark.parametrize(
    "program,basis,keywords",
    [
        pytest.param("cfour", "aug-pvdz", {"frozen_core": False}, id="cfour_conv", marks=using("cfour")),
        pytest.param(
            "gamess", "accd", {"contrl__ispher": 1, "mp2__nacore": 0}, id="gamess_conv", marks=using("gamess")
        ),
        pytest.param(
            "nwchem",
            "aug-cc-pvdz",
            {"basis__spherical": True, "scf__thresh": 1.0e-8, "mp2__freeze": False},
            id="nwchem_conv",
            marks=using("nwchem"),
        ),
        pytest.param(
            "psi4",
            "aug-cc-pvdz",
            {"e_convergence": 1.0e-10, "d_convergence": 1.0e-10, "scf_type": "pk", "mp2_type": "conv"},
            id="psi4_conv",
            marks=using("psi4"),
        ),
    ],
)
@pytest.mark.parametrize(
    "mbe_keywords,anskey,calcinfo_nmbe",
    [
        pytest.param(
            {"bsse_type": "nocp", "return_total_data": False, "max_nbody": 3},
            "nocp_corrected_interaction_energy_through_3_body",
            14,
            id="3b_nocp",
        ),
    ],
)
def test_nbody_he4_single(
    schema_versions, program, basis, keywords, mbe_keywords, anskey, calcinfo_nmbe, he_tetramer, request
):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        from qcmanybody.models.v2 import ManyBodyInput

        AtomicSpecification = models.AtomicSpecification
        subptcl = "cluster_results"
    else:
        from qcmanybody.models.v1 import AtomicSpecification, ManyBodyInput

        subptcl = "component_results"

    # "to_v2"
    subres = "cluster_results" if ("v2" in request.node.name) else "component_results"
    import qcmanybody as qcmb

    if qcmb.__version__.startswith("0.5."):
        subres = "component_results"

    atomic_spec = AtomicSpecification(
        model={"method": "mp2", "basis": basis},
        program=program,
        driver="energy",
        keywords=keywords,
        protocols={"stdout": False},
    )
    mbe_model = ManyBodyInput(
        specification={
            "specification": atomic_spec,
            "keywords": mbe_keywords,
            "driver": "energy",
            "protocols": {subptcl: "all"},
        },
        molecule=he_tetramer,
    )

    mbe_model = checkver_and_convert(mbe_model, request.node.name, "pre")
    ret = qcng.compute(mbe_model, "qcmanybody", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")
    pprint.pprint(ret.model_dump(), width=200)

    assert ret.extras == {}, f"[w] extras wrongly present: {ret.extras.keys()}"

    refs = {
        "nocp_corrected_total_energy_through_1_body": -11.530668717083888,
        "nocp_corrected_total_energy_through_2_body": -11.522851206178828,
        "nocp_corrected_total_energy_through_3_body": -11.523095269671348,
        "nocp_corrected_total_energy_through_4_body": -11.523038093664368,
        "nocp_corrected_interaction_energy_through_1_body": 0.0,
        "nocp_corrected_interaction_energy_through_2_body": 0.007817510905059777,
        "nocp_corrected_interaction_energy_through_3_body": 0.0075734474125397355,
        "nocp_corrected_interaction_energy_through_4_body": 0.007630623419519367,
        "nocp_corrected_2_body_contribution_to_energy": 0.007817510905059777,
        "nocp_corrected_3_body_contribution_to_energy": -0.00024406349252004134,
        "nocp_corrected_4_body_contribution_to_energy": 5.717600697963121e-05,
    }

    ans = refs[anskey]
    ref_nmbe = calcinfo_nmbe
    atol = 1.0e-8

    for skp, ref in refs.items():
        if not "4_body" in skp:
            assert compare_values(ref, getattr(ret.properties, skp), atol=atol, label=f"[b] skprop {skp}")
        else:
            assert getattr(ret.properties, skp) is None

    assert compare_values(
        refs["nocp_corrected_total_energy_through_3_body"],
        ret.properties.nocp_corrected_total_energy,
        atol=atol,
        label=f"[d] skprop tot",
    )
    assert compare_values(
        refs["nocp_corrected_interaction_energy_through_3_body"],
        ret.properties.nocp_corrected_interaction_energy,
        atol=atol,
        label=f"[d] skprop IE",
    )

    assert compare_values(ans, ret.properties.return_energy, atol=atol, label=f"[f] skprop {skp}")
    assert compare_values(ans, ret.return_result, atol=atol, label=f"[g] ret")

    assert ret.properties.calcinfo_nmbe == ref_nmbe, f"[i] {ret.properties.calcinfo_nmbe} != {ref_nmbe}"
    ret_subres = getattr(ret, subres)
    assert len(ret_subres) == ref_nmbe, f"[k] {len(ret_subres)} != {ref_nmbe}; mbe protocol did not take"
    if ref_nmbe > 0:
        an_atres = next(iter(ret_subres.values()))
        assert an_atres.stdout is None, f"[l] atomic protocol did not take"


@uusing("qcmanybody")
@pytest.mark.parametrize(
    "qcprog",
    [
        pytest.param("cfour", marks=using("cfour")),
        pytest.param("gamess", marks=using("gamess")),
        pytest.param("nwchem", marks=using("nwchem")),
        pytest.param("psi4", marks=using("psi4")),
    ],
)
def test_bsse_ene_tu6_cp_ne2(schema_versions, request, qcprog):
    """
    from https://github.com/psi4/psi4/blob/master/tests/tu6-cp-ne2/input.dat
    Example potential energy surface scan and CP-correction for Ne2
    """
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        from qcmanybody.models.v2 import ManyBodyInput
    else:
        from qcmanybody.models.v1 import ManyBodyInput

    tu6_ie_scan = {2.5: 0.757717, 3.0: 0.015685, 4.0: -0.016266}  # Ang: kcal/mol IE

    keywords = {
        "cfour": {"frozen_core": True},
        "gamess": {"contrl__ispher": 1},
        "nwchem": {"ccsd__freeze__atomic": True, "basis__spherical": True},
        "psi4": {"freeze_core": True},
    }
    basis = {
        "cfour": "aug-pvdz",
        "gamess": "accd",
        "nwchem": "aug-cc-pvdz",
        "psi4": "aug-cc-pvdz",
    }

    mbe_data = {
        "specification": {
            "specification": {
                "model": {
                    "method": "ccsd(t)",
                    "basis": basis[qcprog],
                },
                "driver": "energy",
                "program": qcprog,
                "keywords": keywords[qcprog],
            },
            "keywords": {
                "bsse_type": "cp",
            },
            "driver": "energy",
        },
        "molecule": None,
    }

    for R in tu6_ie_scan:
        nene = models.Molecule(
            symbols=["Ne", "Ne"], fragments=[[0], [1]], geometry=[0, 0, 0, 0, 0, R / constants.bohr2angstroms]
        )
        mbe_data["molecule"] = nene

        mbe_model = ManyBodyInput(**mbe_data)
        if qcprog == "gamess":
            with pytest.raises(RuntimeError) as exe:
                qcng.compute(mbe_model, "qcmanybody", raise_error=True, return_version=retver)
            assert "GAMESS+QCEngine can't handle ghost atoms yet" in str(exe.value)
            pytest.xfail("GAMESS can't do ghosts")

        mbe_model = checkver_and_convert(mbe_model, request.node.name, "pre")
        ret = qcng.compute(mbe_model, "qcmanybody", raise_error=True, return_version=retver)
        ret = checkver_and_convert(ret, request.node.name, "post")
        pprint.pprint(ret.model_dump(), width=200)

        assert compare_values(
            tu6_ie_scan[R], ret.return_result * constants.hartree2kcalmol, atol=1.0e-4, label=f"CP-CCSD(T) [{R:3.1f}]"
        )
        assert compare(3, ret.properties.calcinfo_nmbe, label="nmbe")
        assert compare(1, ret.properties.calcinfo_nmc, label="nmc")
        assert compare(2, ret.properties.calcinfo_nfr, label="nfr")
        assert compare(2, ret.properties.calcinfo_natom, label="nat")


@uusing("qcmanybody")
def test_mbe_error(schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        from qcmanybody.models.v2 import ManyBodyInput

        AtomicSpecification = models.AtomicSpecification
    else:
        from qcmanybody.models.v1 import AtomicSpecification, ManyBodyInput

    mbe_data = {
        "specification": {
            "specification": AtomicSpecification(
                **{
                    "model": {
                        "method": "nonsense",
                        "basis": "nonsense",
                    },
                    "driver": "energy",
                    "program": "cms",
                }
            ),
            "keywords": {},
            "driver": "energy",
        },
        "molecule": None,
    }

    nene = models.Molecule(
        symbols=["Ne", "Ne"], fragments=[[0], [1]], geometry=[0, 0, 0, 0, 0, 3.0 / constants.bohr2angstroms]
    )
    mbe_data["molecule"] = nene

    mbe_model = ManyBodyInput(**mbe_data)

    # test 1
    with pytest.raises(RuntimeError) as exc:
        qcng.compute(mbe_model, "qcmanybody", raise_error=True, return_version=retver)

    assert "Program cms is not registered to QCEngine" in str(exc.value)

    # test 2
    ret = qcng.compute(mbe_model, "qcmanybody", return_version=retver)
    assert ret.success is False
    assert "Program cms is not registered to QCEngine" in ret.error.error_message


@uusing("psi4")
@uusing("qcmanybody")
@pytest.mark.parametrize(
    "optimizer,bsse_type,sio",
    [
        pytest.param("optking", "nil", "", marks=using("optking")),
        pytest.param("optking", "cp", False, marks=using("optking_v2")),
        pytest.param("optking", "nocp", False, marks=using("optking_v2")),
        pytest.param("geometric", "nil", "", marks=using("geometric")),
        # needs qcsk v2-compat geometric harness
        # pytest.param("geometric", "nocp", False, marks=using("geometric_genopt")),
        # pytest.param("geometric", "cp", True, marks=using("geometric_genopt")),
    ],
)
def test_bsse_opt_hf_trimer(schema_versions, request, optimizer, bsse_type, sio):
    models, retver, _ = schema_versions

    if bsse_type != "nil" and "as_v2" not in request.node.name:
        pytest.skip(
            "ManyBody Optimization is only available for QCSchema v2. The experimental v1 GeneralizedOptimization is retired."
        )

    Molecule = models.Molecule
    if from_v2(request.node.name):
        from qcmanybody.models.v2 import ManyBodyInput

        subptcl = "cluster_results"
    else:
        from qcmanybody.models.v1 import ManyBodyInput

        subptcl = "component_results"
    # to_v2
    subres = "cluster_results" if ("v2" in request.node.name) else "component_results"

    initial_molecule = Molecule.from_data(
        """
F         -0.04288        2.78905        0.00000
H          0.59079        2.03435        0.00000
--
F         -1.94320       -0.70822        0.00000
H         -1.60642        0.21789       -0.00000
--
F          2.03569       -0.60531       -0.00000
H          1.06527       -0.77673        0.00000
units ang
"""
    )

    at_spec = {
        # schema_name needed for differentiation in genopt
        "schema_name": "qcschema_atomic_specification" if from_v2(request.node.name) else "qcschema_input",
        "program": "psi4",
        "driver": "gradient",
        "model": {
            "method": "hf",
            "basis": "6-31g",
        },
        "keywords": {
            "scf_type": "df",
        },
    }
    if not from_v2(request.node.name):
        at_spec.pop("program")
        at_spec.pop("driver")

    mbe_spec = {
        # schema_name needed for differentiation in genopt
        "schema_name": "qcschema_many_body_specification"
        if from_v2(request.node.name)
        else "qcschema_manybodyspecification",
        "specification": {
            "model": {
                "method": "hf",
                "basis": "6-31g",
            },
            "driver": "energy",
            "program": "psi4",
            "keywords": {},
            "protocols": {
                "stdout": False,
            },
            "extras": {
                "psiapi": True,
            },
        },
        "keywords": {
            "bsse_type": bsse_type,
            "supersystem_ie_only": sio,
        },
        "driver": "energy",
        "protocols": {
            subptcl: "all",
        },
    }

    if from_v2(request.node.name):
        opt_data = {
            "initial_molecule": initial_molecule,
            "specification": {
                "specification": at_spec if (bsse_type == "nil") else mbe_spec,
                "keywords": {
                    "g_convergence": "nwchem_loose",
                },
                "protocols": {
                    "trajectory_results": "initial_and_final",
                },
            },
        }
    else:
        opt_data = {
            "initial_molecule": initial_molecule,
            "input_specification": at_spec if (bsse_type == "nil") else mbe_spec,
            "keywords": {
                "program": "psi4",
                "g_convergence": "nwchem_loose",
            },
            "protocols": {
                "trajectory": "initial_and_final",
            },
        }

    opt_data = checkver_and_convert(opt_data, request.node.name, "pre", cast_dict_as="OptimizationInput")
    ret = qcng.compute(opt_data, optimizer, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    pprint.pprint(ret.model_dump(), width=200)

    r_fh_hb_xptd = {
        "nil": 2.18 / constants.bohr2angstroms,
        "nocp": 2.18 / constants.bohr2angstroms,
        "cp": 2.27 / constants.bohr2angstroms,
    }[bsse_type]
    r_fh_computed = ret.final_molecule.measure([1, 3])
    assert (
        pytest.approx(r_fh_computed, 1.0e-2) == r_fh_hb_xptd
    ), f"hydrogen bond length computed ({r_fh_computed}) != expected ({r_fh_hb_xptd})"
    ret_traj = ret.trajectory_results if "v2" in request.node.name else ret.trajectory
    assert (
        len(ret_traj) == 2
    ), f"trajectory protocol did not take. len(ret.trajectory)={len(ret_traj)} != 2 (initial_and_final)"
    if bsse_type != "nil":
        xptd_nmbe = {
            ("nocp", False): 7,
            ("nocp", True): 4,
            ("cp", False): 10,
            ("cp", True): 7,
        }[(bsse_type, sio)]
        ret_last_subres = getattr(ret_traj[-1], subres)
        assert xptd_nmbe == len(ret_last_subres), f"mbe protocol did not take"
        assert ret_last_subres['["(auto)", [1, 2, 3], [1, 2, 3]]'].stdout is None, f"atomic protocol did not take"


@uusing("qcmanybody")
@pytest.mark.parametrize("schver", [2])  # GeneralizedOptimization retired (this was the v1 precursor)
@pytest.mark.parametrize("bsse_type", ["mbe", "ssfc"])  # aka nocp, cp
@pytest.mark.parametrize(
    "qcprog,qc_keywords",
    [
        pytest.param("psi4", {}, marks=using("psi4")),
        pytest.param("cfour", {}, marks=using("cfour")),
        pytest.param("nwchem", {}, marks=using("nwchem")),
    ],
)
@pytest.mark.parametrize(
    "optimizer,opt_keywords",
    [
        pytest.param("optking", {}, marks=using("optking_v2")),
        # pytest.param(
        #     "geometric", {"convergence_set": "interfrag_tight", "maxiter": 15}, marks=using("geometric_genopt")
        # ),
    ],
)
def test_bsse_opt_lif_dimer(optimizer, opt_keywords, bsse_type, qcprog, qc_keywords, schver):
    # in geomeTRIC: test_lif_bsse

    lif = {
        "symbols": ["Li", "F"],
        "geometry": [0, 0, 0, 0, 0, 3],
        "fragments": [[0], [1]],
        "fragment_charges": [+1, -1],
    }

    mbe_subprop = "component_properties" if schver == 1 else "cluster_properties"
    mbe_subptcl = "component_results" if schver == 1 else "cluster_results"
    opt_subptcl = "trajectory" if schver == 1 else "trajectory_results"

    mbe_spec = {
        "schema_name": "qcschema_many_body_specification" if schver == 2 else "qcschema_manybodyspecification",
        "specification": {
            "model": {
                "method": "hf",
                "basis": "6-31G",
            },
            "driver": "energy",
            "program": qcprog,
            "keywords": qc_keywords,
            "protocols": {
                "stdout": False,
            },
        },
        "keywords": {
            "bsse_type": bsse_type,
            "supersystem_ie_only": True,
        },
        "protocols": {
            mbe_subptcl: "all",
        },
        "driver": "energy",
        "program": "qcmanybody",
    }

    if schver == 2:
        opt_data = {
            "initial_molecule": lif,
            "specification": {
                "specification": at_spec if (bsse_type == "nil") else mbe_spec,
                "keywords": {
                    **opt_keywords,
                },
                "protocols": {
                    opt_subptcl: "final",
                },
            },
        }
    else:
        opt_data = {
            "initial_molecule": lif,
            "input_specification": mbe_spec,
            "keywords": {
                "program": "nonsense",
                **opt_keywords,
            },
            "protocols": {
                opt_subptcl: "final",
            },
        }

    ret = qcng.compute_procedure(opt_data, optimizer, raise_error=True, return_version=schver)

    # printing will show up if job fails
    pprint.pprint(ret.model_dump(), width=200)

    assert ret.success

    ret_traj = ret.trajectory_results if schver == 2 else ret.trajectory
    assert len(getattr(ret_traj[0], mbe_subprop)) == (5 if bsse_type == "ssfc" else 3)

    assert ret.provenance.creator.lower() == optimizer
    assert ret_traj[0].provenance.creator == "QCManyBody"
    atres = list(getattr(ret_traj[0], mbe_subptcl).values())[0]
    assert atres.provenance.creator.lower() == qcprog

    Rlif = ret.final_molecule.measure([0, 1])
    Rref = 3.016 if bsse_type == "ssfc" else 2.969
    assert compare_values(Rlif, Rref, "bond length", atol=1.0e-3)



import pytest
import qcelemental as qcel
from qcelemental.testing import compare_values

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, schema_versions5, using


@pytest.fixture
def h2o_data():
    return """
 # R=0.958 A=104.5
 H                  0.000000000000     1.431430901356     0.984293362719
 O                  0.000000000000     0.000000000000    -0.124038860300
 H                  0.000000000000    -1.431430901356     0.984293362719
 units au
"""


@pytest.fixture
def nh2_data():
    return """
 # R=1.008 #A=105.0
 0 2
 N   0.000000000000000   0.000000000000000  -0.145912918634892
 H   0.000000000000000  -1.511214298139000   1.013682596946108
 H   0.000000000000000   1.511214298139000   1.013682596946108
 units au
 symmetry c1
"""


@pytest.mark.parametrize(
    "program,basis,keywords",
    [
        pytest.param("cfour", "cC-pvdZ", {"scf_conv": 12}, marks=using("cfour")),  # test basis handling, not results
        pytest.param("cfour", "aug-pvdz", {"scf_conv": 12}, marks=using("cfour")),
        pytest.param("cfour", "aug-pvdz", {}, marks=using("cfour")),
        pytest.param(
            "qcore",
            "aug-cc-pVDZ",
            {"coulomb_method": "direct_4idx", "exchange_method": "direct_4idx"},
            marks=using("qcore"),
        ),
        pytest.param("gamess", "accd", {"contrl__ispher": 1}, marks=using("gamess")),
        pytest.param("jaguar", "cc-pVDZ++", {"dconv": 1.0e-8, "econv": 1.0e-9, "nops": 1}, marks=using("jaguar")),
        pytest.param("molpro", "aug-cc-pvdz", {}, marks=using("molpro")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True}, marks=using("nwchem")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True, "qc_module": "tce"}, marks=using("nwchem")),
        pytest.param("psi4", "aug-cc-pvdz", {"scf_type": "direct"}, marks=using("psi4")),
        pytest.param("qchem", "aug-cc-pvdz", {}, marks=using("qchem")),
        pytest.param("turbomole", "aug-cc-pVDZ", {}, marks=using("turbomole")),
        pytest.param("terachem_pbs", "aug-cc-pvdz", {}, marks=using("terachem_pbs")),
    ],
)
def test_sp_hf_rhf_v1v2shim(program, basis, keywords, h2o_data, schema_versions5, request, monkeypatch):
    """cfour/sp-rhf-hf/input.dat
    #! single point HF/adz on water

    """
    models, retver, _ = schema_versions5
    h2o = qcel.models.v2.Molecule.from_data(h2o_data).model_dump()
    if not from_v2(request.node.name):
        h2o["schema_version"] = 2

    # Note that this test is identical to test_sp_hf_rhf for <py314 (but with fewer checks)

    if from_v2(request.node.name):
        resi = {
            "molecule": h2o,
            "specification": {"driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords},
        }
    else:
        resi = {"molecule": h2o, "driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords}

    monkeypatch.setenv("QCNG_USE_V1V2_SHIM", "1")
    # resi = checkver_and_convert(resi, request.node.name, "pre")
    res = qcng.compute(resi, program, raise_error=True, return_dict=True, return_version=retver)
    import pprint

    pprint.pprint(res, width=200)
    # res = checkver_and_convert(res, request.node.name, "post")

    if "_v2" in request.node.name:
        assert res["input_data"]["specification"]["driver"] == "energy"
    else:
        assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    if basis == "cC-pvdZ":
        return

    # aug-cc-pvdz
    scf_tot = -76.0413815332

    atol = 1.0e-6
    assert compare_values(scf_tot, res["return_result"], atol=atol)


@pytest.mark.parametrize(
    "program,basis,keywords",
    [
        pytest.param("cfour", "cC-pvdZ", {"scf_conv": 12}, marks=using("cfour")),  # test basis handling, not results
        pytest.param("cfour", "aug-pvdz", {"scf_conv": 12}, marks=using("cfour")),
        pytest.param("cfour", "aug-pvdz", {}, marks=using("cfour")),
        pytest.param(
            "qcore",
            "aug-cc-pVDZ",
            {"coulomb_method": "direct_4idx", "exchange_method": "direct_4idx"},
            marks=using("qcore"),
        ),
        pytest.param("gamess", "accd", {"contrl__ispher": 1}, marks=using("gamess")),
        pytest.param("jaguar", "cc-pVDZ++", {"dconv": 1.0e-8, "econv": 1.0e-9, "nops": 1}, marks=using("jaguar")),
        pytest.param("molpro", "aug-cc-pvdz", {}, marks=using("molpro")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True}, marks=using("nwchem")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True, "qc_module": "tce"}, marks=using("nwchem")),
        pytest.param("psi4", "aug-cc-pvdz", {"scf_type": "direct"}, marks=using("psi4")),
        pytest.param("qchem", "aug-cc-pvdz", {}, marks=using("qchem")),
        pytest.param("turbomole", "aug-cc-pVDZ", {}, marks=using("turbomole")),
        pytest.param("terachem_pbs", "aug-cc-pvdz", {}, marks=using("terachem_pbs")),
    ],
)
def test_sp_hf_rhf(program, basis, keywords, h2o_data, schema_versions, request):
    """cfour/sp-rhf-hf/input.dat
    #! single point HF/adz on water

    """
    models, retver, _ = schema_versions
    h2o = models.Molecule.from_data(h2o_data)

    if from_v2(request.node.name):
        resi = {
            "molecule": h2o,
            "specification": {"driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords},
        }
    else:
        resi = {"molecule": h2o, "driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords}

    resi = checkver_and_convert(resi, request.node.name, "pre")
    res = qcng.compute(resi, program, raise_error=True, return_dict=True, return_version=retver)
    res = checkver_and_convert(res, request.node.name, "post")

    if "v2" in request.node.name:
        assert res["input_data"]["specification"]["driver"] == "energy"
    else:
        assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    if basis == "cC-pvdZ":
        return

    # aug-cc-pvdz
    scf_tot = -76.0413815332

    atol = 1.0e-6
    assert compare_values(scf_tot, res["return_result"], atol=atol)


@pytest.mark.parametrize(
    "program,basis,keywords",
    [
        pytest.param(
            "cfour",
            "aug-pvdz",
            {"reference": "uhf", "occupation": [[3, 1, 1, 0], [3, 0, 1, 0]], "scf_conv": 12},
            marks=using("cfour"),
        ),
        pytest.param("cfour", "aug-pvdz", {"reference": "uhf"}, marks=using("cfour")),
        pytest.param(
            "qcore",
            "aug-cc-pVDZ",
            {"ansatz": "u", "coulomb_method": "direct_4idx", "exchange_method": "direct_4idx"},
            marks=using("qcore"),
        ),
        pytest.param("gamess", "accd", {"contrl__ispher": 1, "contrl__scftyp": "uhf"}, marks=using("gamess")),
        pytest.param(
            "jaguar", "cc-pVDZ++", {"dconv": 1.0e-8, "econv": 1.0e-9, "iuhf": 1, "nops": 1}, marks=using("jaguar")
        ),
        pytest.param("molpro", "aug-cc-pvdz", {"reference": "unrestricted"}, marks=using("molpro")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True, "scf__uhf": True}, marks=using("nwchem")),
        pytest.param(
            "nwchem",
            "aug-cc-pvdz",
            {"basis__spherical": True, "qc_module": "tce", "scf__uhf": True},
            marks=using("nwchem"),
        ),
        pytest.param("psi4", "aug-cc-pvdz", {"reference": "uhf", "scf_type": "direct"}, marks=using("psi4")),
        pytest.param("qchem", "aug-cc-pvdz", {}, marks=using("qchem")),
        pytest.param("turbomole", "aug-cc-pVDZ", {}, marks=using("turbomole")),
    ],
)
def test_sp_hf_uhf(program, basis, keywords, nh2_data, schema_versions, request):
    models, retver, _ = schema_versions

    nh2 = models.Molecule.from_data(nh2_data)
    if from_v2(request.node.name):
        resi = {
            "molecule": nh2,
            "specification": {"driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords},
        }
    else:
        resi = {"molecule": nh2, "driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords}
    resi = models.AtomicInput(**resi)

    resi = checkver_and_convert(resi, request.node.name, "pre")
    res = qcng.compute(resi, program, raise_error=True, return_dict=False, return_version=retver)
    res = checkver_and_convert(res, request.node.name, "post")

    assert res.success is True
    res = res.model_dump()
    if "v2" in request.node.name:
        assert res["input_data"]["specification"]["driver"] == "energy"
    else:
        assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    # aug-cc-pvdz
    scf_tot = -55.57513805253009

    atol = 1.0e-6
    assert compare_values(scf_tot, res["return_result"], atol=atol)


@pytest.mark.parametrize(
    "program,basis,keywords",
    [
        pytest.param(
            "cfour",
            "aug-pvdz",
            {"reference": "rohf", "occupation": [[3, 1, 1, 0], [3, 0, 1, 0]], "scf_conv": 12},
            marks=using("cfour"),
        ),
        pytest.param("cfour", "aug-pvdz", {"reference": "rohf"}, marks=using("cfour")),
        pytest.param("gamess", "accd", {"contrl__ispher": 1, "contrl__scftyp": "rohf"}, marks=using("gamess")),
        pytest.param(
            "jaguar", "cc-pVDZ++", {"dconv": 1.0e-8, "econv": 1.0e-9, "iuhf": 0, "nops": 1}, marks=using("jaguar")
        ),
        pytest.param("molpro", "aug-cc-pvdz", {}, marks=using("molpro")),
        pytest.param("nwchem", "aug-cc-pvdz", {"basis__spherical": True, "scf__rohf": True}, marks=using("nwchem")),
        pytest.param(
            "nwchem",
            "aug-cc-pvdz",
            {"basis__spherical": True, "qc_module": "tce", "scf__rohf": True},
            marks=using("nwchem"),
        ),
        pytest.param("psi4", "aug-cc-pvdz", {"reference": "rohf", "scf_type": "direct"}, marks=using("psi4")),
        pytest.param("qchem", "aug-cc-pvdz", {"UNRESTRICTED": False}, marks=using("qchem")),
    ],
)
def test_sp_hf_rohf(program, basis, keywords, nh2_data, schema_versions, request):
    models, retver, models_out = schema_versions

    nh2 = models.Molecule.from_data(nh2_data)
    if from_v2(request.node.name):
        resi = {
            "molecule": nh2,
            "specification": {"driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords},
        }
    else:
        resi = {"molecule": nh2, "driver": "energy", "model": {"method": "hf", "basis": basis}, "keywords": keywords}

    resi = checkver_and_convert(resi, request.node.name, "pre")
    res = qcng.compute(resi, program, raise_error=True, return_dict=True, return_version=retver)
    res = checkver_and_convert(res, request.node.name, "post")

    if "v2" in request.node.name:
        assert res["input_data"]["specification"]["driver"] == "energy"
    else:
        assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    # aug-cc-pvdz
    scf_tot = -55.570724348574

    atol = 1.0e-6
    assert compare_values(scf_tot, res["return_result"], atol=atol)



"""QCEngine harness for Jaguar.

The Schrödinger Python modules are intentionally imported at run time so that
QCEngine itself does not depend on the Schrödinger suite.
"""

import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional

import numpy as np
from qcelemental import constants
from qcelemental.models.v2 import AtomicResult, BasisSet, Provenance
from qcelemental.util import safe_version, which_import

from ..exceptions import InputError, UnknownError
from ..util import environ_context, temporary_directory
from .model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models.v2 import AtomicInput

    from ..config import TaskConfig


class JaguarHarness(ProgramHarness):
    """Run Jaguar calculations through the Schrödinger Python API.

    Jaguar is file based, so the harness uses a dedicated scratch directory.
    The current launch path temporarily changes process-wide state and is
    therefore not thread-safe. A single calculation can use multiple threads
    through Jaguar's ``-PARALLEL`` option, but the harness does not configure
    multi-node execution. QCEngine does not currently configure Jaguar's
    memory limit.
    """

    _defaults: ClassVar[Dict[str, Any]] = {
        "name": "Jaguar",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """Check whether the Schrödinger Jaguar Python API is importable.

        Parameters
        ----------
        raise_error
            Raise an informative import error instead of returning ``False``
            when the API is unavailable.

        Returns
        -------
        bool
            Whether the Jaguar input module can be imported.
        """
        return which_import(
            "schrodinger.application.jaguar.input",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Run QCEngine from a Schrodinger Python environment with SCHRODINGER defined.",
        )

    def get_version(self) -> str:
        """Return the normalized Jaguar product version.

        Schrödinger stores this as an integer such as ``134039``, denoting
        Jaguar 13.4 release 39. The normalized value is cached by the path of
        the module that supplies the product version.

        Returns
        -------
        str
            A PEP 440-compatible Jaguar version such as ``13.4.39``.
        """
        self.found(raise_error=True)

        module_path = which_import("schrodinger.infra.mm")
        cache_key = f"jaguar:{module_path}"
        if cache_key not in self.version_cache:
            from schrodinger.infra import mm

            encoded_version = int(mm.mmfile_get_product_version("jaguar"))
            major_minor, release = divmod(encoded_version, 1000)
            major, minor = divmod(major_minor, 10)
            self.version_cache[cache_key] = safe_version(f"{major}.{minor}.{release:03d}")
        return self.version_cache[cache_key]

    def get_suite_version(self) -> str:
        """Return the normalized Schrödinger suite release.

        Returns
        -------
        str
            A PEP 440-compatible suite release such as ``2026.post4``.
        """
        self.found(raise_error=True)

        module_path = which_import("schrodinger.application.jaguar.constants")
        cache_key = f"suite:{module_path}"
        if cache_key not in self.version_cache:
            from schrodinger.application.jaguar.constants import VERSION

            self.version_cache[cache_key] = safe_version(VERSION)
        return self.version_cache[cache_key]

    @staticmethod
    def _validate_input(input_model: "AtomicInput") -> None:
        """Validate that an atomic input uses features supported by Jaguar.

        Parameters
        ----------
        input_model
            The QCSchema atomic input to validate.

        Raises
        ------
        InputError
            If the driver is unsupported, the basis is an explicit QCSchema
            ``BasisSet``, or the optional
            Jaguar guess input is malformed.
        """
        driver = input_model.specification.driver
        if driver not in {"energy", "gradient", "hessian"}:
            raise InputError(f"Driver {driver} not implemented for Jaguar.")

        basis = input_model.specification.model.basis
        if isinstance(basis, BasisSet):
            raise InputError("QCSchema BasisSet for model.basis not implemented. Use a string basis name.")

        JaguarHarness._get_guess_input(input_model)

    @staticmethod
    def _get_guess_input(input_model: "AtomicInput") -> Optional[str]:
        """Return and validate the optional Jaguar guess input text.

        Parameters
        ----------
        input_model
            Atomic input whose ``specification.extras.jaguar`` mapping may
            contain a ``guess_input`` string.

        Returns
        -------
        str or None
            Text of a Jaguar input containing an ``&guess`` section, or
            ``None`` when no guess was supplied.

        Raises
        ------
        InputError
            If the Jaguar extras or ``guess_input`` have invalid types, or
            the guess string is empty.
        """
        jaguar_extras = input_model.specification.extras.get("jaguar", {})
        if not isinstance(jaguar_extras, dict):
            raise InputError("specification.extras.jaguar must be a mapping.")

        guess_input = jaguar_extras.get("guess_input")
        if guess_input is None:
            return None
        if not isinstance(guess_input, str) or not guess_input.strip():
            raise InputError("specification.extras.jaguar.guess_input must be a non-empty string.")
        return guess_input

    @staticmethod
    def _append_guess_sections(jaguar_input: Any, guess_input: str, job_name: str) -> None:
        """Copy guess-related sections from text into a Jaguar input object.

        Parameters
        ----------
        jaguar_input
            Target ``JaguarInput`` configured from the current QCSchema
            molecule, model, driver, and keywords.
        guess_input
            Contents of a Jaguar input file containing an ``&guess`` section
            and, optionally, an ``&guess_basis`` section.
        job_name
            Target job path, used to place a temporary source input in the
            same scratch directory.

        Raises
        ------
        InputError
            If Jaguar cannot parse the supplied text or it has no ``&guess``
            section.
        """
        from schrodinger.application.jaguar.input import JaguarInput
        from schrodinger.infra import mm

        guess_file = Path(job_name).parent / "_qcengine_guess.in"
        try:
            guess_file.write_text(guess_input, encoding="utf-8")
            source_input = JaguarInput(str(guess_file), compute_connectivity=False)
            if not source_input.sectionDefined("guess"):
                raise InputError("specification.extras.jaguar.guess_input has no &guess section.")

            for section in ("guess", "guess_basis"):
                if source_input.sectionDefined(section):
                    section_body = source_input.getSectionText(section)
                    if section_body is not None:
                        mm.mmjag_sect_append_wrapper(jaguar_input.handle, f"&{section}\n{section_body}\n&\n")
        except InputError:
            raise
        except Exception as exc:
            raise InputError(f"Could not parse Jaguar guess_input: {exc}") from exc

    @staticmethod
    def _build_jaguar_input(input_model: "AtomicInput", job_name: str):
        """Translate a QCSchema atomic input into a Jaguar input object.

        Parameters
        ----------
        input_model
            The validated atomic input containing the molecule, model,
            driver, and native Jaguar keywords.
        job_name
            Basename or path to use for the Jaguar job and input file.

        Returns
        -------
        schrodinger.application.jaguar.input.JaguarInput
            A configured Jaguar input ready to be saved or run.

        Notes
        -----
        QCSchema coordinates are converted from bohr to angstrom. Model and
        driver settings override conflicting values in the native keywords.
        Atoms marked non-real by QCSchema retain their element and are marked as
        counterpoise atoms, which Jaguar serializes by appending an at sign to
        the atom label.
        If ``specification.extras.jaguar.guess_input`` is present, only its
        ``&guess`` and optional ``&guess_basis`` sections are copied; its old
        geometry and ``&gen`` settings are intentionally ignored.
        """
        from schrodinger import structure
        from schrodinger.application.jaguar.input import JaguarInput

        molecule = input_model.molecule
        geometry = np.asarray(molecule.geometry) * constants.bohr2angstroms
        jaguar_structure = structure.create_new_structure()
        for symbol, xyz in zip(molecule.symbols, geometry):
            jaguar_structure.addAtom(str(symbol), *map(float, xyz))

        keywords = {key.lower(): value for key, value in input_model.specification.keywords.items()}
        model = input_model.specification.model

        # Model and driver fields are authoritative over duplicate native keywords.
        keywords["dftname"] = "HF" if model.method.lower() in {"hf", "scf"} else model.method
        if model.basis:
            keywords["basis"] = model.basis

        driver = input_model.specification.driver
        keywords["igeopt"] = -1 if driver == "gradient" else 0
        keywords["ifreq"] = 1 if driver == "hessian" else 0
        keywords["isymm"] = 0

        jaguar_input = JaguarInput(name=job_name, structure=jaguar_structure, genkeys=keywords)
        guess_input = JaguarHarness._get_guess_input(input_model)
        if guess_input is not None:
            JaguarHarness._append_guess_sections(jaguar_input, guess_input, job_name)

        for atom_index, real in enumerate(molecule.real, start=1):
            if not real:
                jaguar_input._setCounterpoise(atom_index, True)
        jaguar_input.setValue("molchg", int(molecule.molecular_charge))
        jaguar_input.setValue("multip", molecule.molecular_multiplicity)
        return jaguar_input

    @staticmethod
    def _optional_properties(jaguar_output: Any, input_model: "AtomicInput") -> Dict[str, Any]:
        """Translate available Jaguar results into QCSchema properties.

        Parameters
        ----------
        jaguar_output
            Parsed ``JaguarOutput`` returned by the Jaguar calculation.
        input_model
            The atomic input, used for molecule-derived metadata.

        Returns
        -------
        dict
            QCSchema ``AtomicResultProperties`` fields available in the
            Jaguar output. Missing optional quantities are omitted.
        """
        results = jaguar_output.last_results
        properties: Dict[str, Any] = {
            "calcinfo_natom": len(input_model.molecule.symbols),
            "return_energy": results.energy,
        }

        direct_mapping = {
            "nbasis": "calcinfo_nbasis",
            "nuclear_repulsion": "nuclear_repulsion_energy",
            "scf_energy": "scf_total_energy",
            "energy_one_electron": "scf_one_electron_energy",
            "energy_two_electron": "scf_two_electron_energy",
            "rimp2_ss_energy": "mp2_same_spin_correlation_energy",
            "rimp2_os_energy": "mp2_opposite_spin_correlation_energy",
            "rimp2_corr_energy": "mp2_correlation_energy",
            "rimp2_energy": "mp2_total_energy",
        }
        for source, target in direct_mapping.items():
            owner = jaguar_output if source == "nbasis" else results
            value = getattr(owner, source, None)
            if value is not None:
                properties[target] = value
        if "calcinfo_nbasis" in properties:
            properties["calcinfo_nmo"] = properties["calcinfo_nbasis"]

        nelectron = getattr(jaguar_output, "nelectron", None)
        if nelectron is not None:
            spin = int(input_model.molecule.molecular_multiplicity) - 1
            nalpha = (nelectron + spin) // 2
            nbeta = (nelectron - spin) // 2
        else:
            nalpha = getattr(jaguar_output, "num_occ_orbs_alpha", None)
            nbeta = getattr(jaguar_output, "num_occ_orbs_beta", None)
            if nalpha is None and nbeta is None:
                nocc = getattr(jaguar_output, "num_occ_orbs", None)
                nalpha = nbeta = nocc
        if nalpha is not None:
            properties["calcinfo_nalpha"] = nalpha
        if nbeta is not None:
            properties["calcinfo_nbeta"] = nbeta

        dipole = getattr(results, "dipole_qm", None)
        if dipole is not None and None not in (dipole.x, dipole.y, dipole.z):
            properties["scf_dipole_moment"] = np.asarray([dipole.x, dipole.y, dipole.z]) / constants.dipmom_au2debye

        return properties

    @staticmethod
    def _qcvars(properties: Dict[str, Any], driver: str, method: str) -> Dict[str, Any]:
        """Build legacy QCVariable aliases required by QCEngine's standard suite.

        QCSchema properties remain the canonical result representation. The
        aliases here let older cross-program tests and clients query the same
        Hartree--Fock results through their traditional QCVariable names.
        """
        qcvars = {"CURRENT ENERGY": properties["return_energy"]}
        if method.lower() != "hf":
            return qcvars

        scf_energy = properties["scf_total_energy"]
        qcvars.update(
            {
                "HF TOTAL ENERGY": scf_energy,
                "SCF TOTAL ENERGY": scf_energy,
                "CURRENT REFERENCE ENERGY": scf_energy,
            }
        )
        if driver == "gradient":
            gradient = properties["return_gradient"]
            qcvars.update({"HF TOTAL GRADIENT": gradient, "CURRENT GRADIENT": gradient})
        elif driver == "hessian":
            hessian = properties["return_hessian"]
            qcvars.update({"HF TOTAL HESSIAN": hessian, "CURRENT HESSIAN": hessian})

        return qcvars

    @staticmethod
    def _read_hessian(job_base: Path, natom: int):
        """Read a Cartesian Hessian from a generated Jaguar input file.

        Parameters
        ----------
        job_base
            Path to the Jaguar job without a filename extension.
        natom
            Number of atoms, used to validate the Hessian dimensions.

        Returns
        -------
        tuple[numpy.ndarray, pathlib.Path]
            The ``(3 * natom, 3 * natom)`` Hessian and the generated input
            file containing its ``&hess`` section.

        Raises
        ------
        UnknownError
            If no Hessian is found or its dimensions are inconsistent.
        """
        from schrodinger.application.jaguar.input import JaguarInput

        expected_shape = (3 * natom, 3 * natom)
        candidates = [job_base.with_suffix(".01.in"), *sorted(job_base.parent.glob(f"{job_base.name}*.01.in"))]
        for candidate in dict.fromkeys(candidates):
            if not candidate.is_file():
                continue
            hessian = JaguarInput(str(candidate), compute_connectivity=False).getHessian()
            if hessian is not None:
                hessian = np.asarray(hessian)
                if hessian.shape != expected_shape:
                    raise UnknownError(
                        f"Jaguar Hessian in {candidate.name} has shape {hessian.shape}; expected {expected_shape}."
                    )
                return hessian, candidate

        raise UnknownError("Jaguar completed the Hessian calculation but no &hess section was found.")

    @staticmethod
    def _collect_native_files(job_base: Path, restart_file: Optional[Path]) -> Dict[str, str]:
        """Read selected Jaguar files before the scratch directory is removed.

        Parameters
        ----------
        job_base
            Path to the Jaguar job without a filename extension.
        restart_file
            Generated input known to contain the requested Hessian, if any.

        Returns
        -------
        dict
            Text contents of the original input, generated numbered inputs,
            and Jaguar log, keyed by their native filenames. QCSchema output
            protocols determine which entries survive in the final result.
        """
        native_files = {}
        input_file = job_base.with_suffix(".in")
        if input_file.is_file():
            native_files["input"] = input_file.read_text(errors="replace")

        generated_inputs = [restart_file, *sorted(job_base.parent.glob(f"{job_base.name}.*.in"))]
        for generated_input in dict.fromkeys(generated_inputs):
            if generated_input is not None and generated_input.is_file():
                native_files[generated_input.name] = generated_input.read_text(errors="replace")

        log_file = job_base.with_suffix(".log")
        if log_file.is_file():
            native_files[log_file.name] = log_file.read_text(errors="replace")
        return native_files

    def compute(self, input_model: "AtomicInput", config: "TaskConfig") -> AtomicResult:
        """Execute Jaguar and construct a QCSchema atomic result.

        Parameters
        ----------
        input_model
            Atomic input specifying the molecule, model, driver, keywords,
            and output protocols.
        config
            QCEngine task resources and scratch-directory configuration.

        Returns
        -------
        AtomicResult
            Validated QCSchema result containing the requested energy,
            gradient, or Hessian and available auxiliary properties.

        Raises
        ------
        InputError
            If the requested input feature is unsupported.
        UnknownError
            If Jaguar fails, omits the requested result, or produces an
            inconsistent Hessian.
        """
        self.found(raise_error=True)
        self._validate_input(input_model)

        with temporary_directory(
            parent=config.scratch_directory, suffix="_jaguar_scratch", messy=config.scratch_messy
        ) as tmpdir:
            job_base = Path(tmpdir) / "dispatch"
            jaguar_input = self._build_jaguar_input(input_model, str(job_base))

            previous_directory = Path.cwd()
            try:
                # Job Control returns result files to the launch directory.
                os.chdir(tmpdir)
                with environ_context(config=config):
                    jaguar_output = jaguar_input.run(**{"-PARALLEL": str(config.ncores)})
            except Exception as exc:
                raise UnknownError(str(exc)) from exc
            finally:
                os.chdir(previous_directory)

            if jaguar_output.status != jaguar_output.OK:
                raise UnknownError(
                    f"Jaguar calculation failed with ERROR {jaguar_output.fatal_errorno}: "
                    f"{jaguar_output.fatal_error}"
                )

            properties = self._optional_properties(jaguar_output, input_model)
            results = jaguar_output.last_results
            driver = input_model.specification.driver
            restart_file = None

            if driver == "energy":
                return_result = results.energy
            elif driver == "gradient":
                if results.forces is None:
                    raise UnknownError("Jaguar completed the gradient calculation but returned no forces.")
                return_result = -np.asarray(results.forces)
                properties["return_gradient"] = return_result
                if input_model.specification.model.method.lower() == "hf":
                    properties["scf_total_gradient"] = return_result
            else:
                return_result, restart_file = self._read_hessian(job_base, len(input_model.molecule.symbols))
                if input_model.specification.model.method.lower() == "hf":
                    properties["scf_total_hessian"] = return_result
                properties["return_hessian"] = return_result

            output_file = job_base.with_suffix(".out")
            stdout = output_file.read_text(errors="replace") if output_file.is_file() else None
            native_files = self._collect_native_files(job_base, restart_file)

        provenance = Provenance(
            creator="Jaguar",
            version=self.get_version(),
            routine="schrodinger.application.jaguar.input.JaguarInput.run",
            nthreads=config.ncores,
        )
        return AtomicResult(
            input_data=input_model,
            molecule=input_model.molecule,
            properties=properties,
            return_result=return_result,
            stdout=stdout,
            native_files=native_files,
            success=True,
            provenance=provenance,
            extras={
                "qcvars": self._qcvars(properties, driver, input_model.specification.model.method),
                "jaguar": {
                    "suite_version": self.get_suite_version(),
                    "point_group": jaguar_output.point_group,
                },
            },
        )



from typing import TYPE_CHECKING, Any, ClassVar, Dict

from qcelemental.models.v2 import AtomicResult, Provenance
from qcelemental.util import safe_version, which_import

from qcengine.exceptions import InputError
from qcengine.programs.model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models.v2 import AtomicInput

    from qcengine.config import TaskConfig


class AIMNET2Harness(ProgramHarness):
    """A harness to run AIMNET2 models <https://github.com/isayevlab/AIMNet2>"""

    _CACHE = {}

    _defaults: ClassVar[Dict[str, Any]] = {
        "name": "AIMNET2",
        "scratch": False,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }

    version_cache: Dict[str, str] = {}

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which_import(
            "pyaimnet2",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `pip install git+https://github.com/jthorton/AIMNet2.git@main`",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("pyaimnet2")
        if which_prog not in self.version_cache:
            import pyaimnet2

            self.version_cache[which_prog] = safe_version(pyaimnet2.__version__)

        return self.version_cache[which_prog]

    def load_model(self, name: str):
        model_name = name.lower()
        if model_name in self._CACHE:
            return self._CACHE[model_name]

        from pyaimnet2 import load_model

        model = load_model(model_name=model_name)
        self._CACHE[model_name] = model
        return self._CACHE[model_name]

    def compute(self, input_data: "AtomicInput", config: "TaskConfig"):
        self.found(raise_error=True)
        import torch

        from qcengine.units import ureg

        # check we can run on the set of elements
        known_elements = {"H", "B", "C", "N", "O", "F", "Si", "P", "S", "Cl", "As", "Se", "Br", "I"}
        target_elements = set(input_data.molecule.symbols)

        unknown_elements = target_elements - known_elements
        if unknown_elements:
            raise InputError(
                f"AIMNET2 model {input_data.specification.model.method} does not support elements {unknown_elements}."
            )

        method = input_data.specification.model.method
        # load the model using the method as the file name
        model = self.load_model(name=method)

        # build the required input data
        aimnet_input = {
            "coord": torch.tensor(
                [input_data.molecule.geometry * ureg.conversion_factor("bohr", "angstrom")],
                dtype=torch.float64,
                device="cpu",
            ),
            "numbers": torch.tensor([input_data.molecule.atomic_numbers], dtype=torch.long, device="cpu"),
            "charge": torch.tensor([input_data.molecule.molecular_charge], dtype=torch.float64, device="cpu"),
        }

        if input_data.specification.driver == "gradient":
            aimnet_input["coord"].requires_grad_(True)
        out = model(aimnet_input)

        ret_data = {
            "input_data": input_data,
            "molecule": input_data.molecule,
            "success": False,
            "properties": {
                "return_energy": out["energy"].item() * ureg.conversion_factor("eV", "hartree"),
                "return_gradient": (
                    -1.0 * out["forces"][0].detach().numpy() * ureg.conversion_factor("eV / angstrom", "hartree / bohr")
                ),
                "calcinfo_natom": len(input_data.molecule.atomic_numbers),
            },
            "extras": {},
        }
        # update with calculated extras
        ret_data["extras"]["aimnet2"] = {
            "charges": out["charges"].detach()[0].cpu().numpy(),
            "ensemble_charges_std": out["charges_std"].detach()[0].cpu().numpy(),
            "ensemble_energy_std": out["energy_std"].item(),
            "ensemble_forces_std": out["forces_std"].detach()[0].cpu().numpy(),
        }
        if input_data.specification.driver == "energy":
            ret_data["return_result"] = ret_data["properties"]["return_energy"]
        elif input_data.specification.driver == "gradient":
            ret_data["return_result"] = ret_data["properties"]["return_gradient"]
        else:
            raise InputError(
                f"AIMNET2 can only compute energy and gradients driver methods. Requested {input_data.specification.driver} not supported."
            )

        ret_data["provenance"] = Provenance(creator="pyaimnet2", version=self.get_version(), routine="load_model")

        ret_data["success"] = True

        return AtomicResult(**ret_data)



import logging
from io import StringIO
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Union

from qcelemental.util import parse_version, safe_version, which_import

from ..exceptions import InputError, RandomError, ResourceError, UnknownError
from .model import ProcedureHarness

if TYPE_CHECKING:
    from ..config import TaskConfig


def _handle_errors(output_data: Dict[str, Any], default_error: str) -> tuple[str, str]:
    error_block = output_data.get("error")
    if isinstance(error_block, dict):
        if "error_message" in error_block:
            return error_block["error_message"], error_block.get("error_type", "unknown_error")
        return str(error_block), "unknown_error"
    if isinstance(error_block, str):
        return error_block, "unknown_error"
    return default_error, "unknown_error"


class OptKingProcedure(ProcedureHarness):

    _defaults: ClassVar[Dict[str, Any]] = {"name": "OptKing", "procedure": "optimization"}

    version_cache: Dict[str, str] = {}

    def found(self, raise_error: bool = False) -> bool:
        return which_import(
            "optking",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install optking -c conda-forge`.",
        )

    def build_input_model(
        self, data: Union[Dict[str, Any], "OptimizationInput"], *, return_input_schema_version: bool = False
    ) -> "OptimizationInput":
        return self._build_model(data, "OptimizationInput", return_input_schema_version=return_input_schema_version)

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("optking")
        if which_prog not in self.version_cache:
            import optking

            self.version_cache[which_prog] = safe_version(optking.__version__)

        return self.version_cache[which_prog]

    def compute(self, input_model: "OptimizationInput", config: "TaskConfig") -> "OptimizationResult":
        self.found(raise_error=True)
        import optking

        log_stream = StringIO()
        logname = f"{optking.log_name}{optking.__name__}"
        log = logging.getLogger(logname)
        log.addHandler(logging.StreamHandler(log_stream))
        log.setLevel("INFO")

        if parse_version(self.get_version()) < parse_version("0.5"):
            from qcelemental.models.v1 import OptimizationResult

            input_data_v1 = input_model.convert_v(1).dict()

            # Set retries to two if zero while respecting local_config
            local_config = config.dict()
            local_config["retries"] = local_config.get("retries", 2) or 2
            input_data_v1["input_specification"]["extras"]["_qcengine_local_config"] = local_config

            # Run the program
            output_v1 = optking.optwrapper.optimize_qcengine(input_data_v1)
            output_v1["stdout"] = log_stream.getvalue()

            output_v1["input_specification"]["extras"].pop("_qcengine_local_config", None)
            if output_v1["success"]:
                output_v1 = OptimizationResult(**output_v1)
                output = output_v1.convert_v(2, external_input_data=input_model)
            else:
                error_message, error_type = _handle_errors(output_v1, "OptKing failed without an error message.")
                error_cls = {
                    "random_error": RandomError,
                    "unknown_error": UnknownError,
                    "input_error": InputError,
                    "resource_error": ResourceError,
                }.get(error_type, UnknownError)
                raise error_cls(
                    error_message,
                    extras={
                        "failed_result": output_v1,
                    },
                )

        else:
            # optking v0.5 can run QCSchema 1<->1 and 2<->2, so update to 2
            from qcelemental.models.v2 import OptimizationResult

            input_data_v2 = input_model.model_dump()

            # Set retries to two if zero while respecting local_config
            local_config = config.model_dump()
            local_config["retries"] = local_config.get("retries", 2) or 2
            input_data_v2["specification"]["specification"]["extras"]["_qcengine_local_config"] = local_config

            # Run the program
            output_v2 = optking.optwrapper.optimize_qcengine(input_data_v2)
            output_v2["stdout"] = log_stream.getvalue()

            output_v2["input_data"]["specification"]["specification"]["extras"].pop("_qcengine_local_config", None)
            if output_v2["success"]:
                output = OptimizationResult(**output_v2)
            else:
                error_message, error_type = _handle_errors(output_v2, "OptKing failed without an error message.")
                error_cls = {
                    "random_error": RandomError,
                    "unknown_error": UnknownError,
                    "input_error": InputError,
                    "resource_error": ResourceError,
                }.get(error_type, UnknownError)
                raise error_cls(
                    error_message,
                    extras={
                        "failed_result": output_v2,
                    },
                )

        return output



import json
import os
import subprocess
import sys
from typing import List

import pytest

from qcengine import cli, get_molecule, util
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, uusing


def run_qcengine_cli(args: List[str], stdin: str = None) -> str:
    """
    Runs qcengine via its CLI

    This method was chosen over the more sophisticated util.execute in order to get pytest-cov to work.
    For the same reason, qcengine is invoked as "python -m qcengine.cli" rather than "qcengine".

    Parameters
    ----------
    args: List[str]
        List of CLI arguments.
    stdin: Optional[str]
        Standard input for the process.

    Returns
    -------
    str
        QCEngine CLI standard output.
    """
    if stdin is not None:
        stdin = stdin.encode("utf-8")

    return subprocess.check_output([sys.executable, "-m", "qcengine"] + args, input=stdin)


def test_no_args():
    """Test for qcengine with no arguments"""
    try:
        run_qcengine_cli([])
    except subprocess.CalledProcessError as e:
        assert e.returncode == 1


@pytest.mark.skipif(
    sys.version_info >= (3, 14), reason="too many programs throw errors on import with Py >=3.14. revisit soon!"
)
def test_info():
    """Test for qcengine info"""
    outputs = []
    for arg in cli.info_choices:
        output = run_qcengine_cli(["info", arg])
        if arg not in {"all", "config"}:  # output of config changes call-to-call depending e.g. on mem available
            outputs.append(output)

    default_output = run_qcengine_cli(["info"])
    for output in outputs:
        assert output in default_output


@uusing("psi4")
def test_run_psi4(tmp_path, schema_versions, request):
    """Tests qcengine run with psi4 and JSON input"""
    models, retver, _ = schema_versions

    def check_result(stdout):
        output = json.loads(stdout)
        output = checkver_and_convert(output, request.node.name, "post")
        assert output["provenance"]["creator"].lower() == "psi4"
        assert output["success"] is True

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=models.Molecule(**get_molecule("hydrogen", return_dict=True)),
            specification=models.AtomicSpecification(
                driver="energy",
                model={"method": "hf", "basis": "6-31G"},
            ),
        )
    else:
        inp = models.AtomicInput(
            molecule=models.Molecule(**get_molecule("hydrogen", return_dict=True)),
            driver="energy",
            model={"method": "hf", "basis": "6-31G"},
        )

    inp = checkver_and_convert(inp, request.node.name, "pre")
    args = ["run", "psi4", f"--return-version={retver}", inp.model_dump_json()]
    check_result(run_qcengine_cli(args))

    args = ["run", "psi4", os.path.join(tmp_path, "input.json"), f"--return-version={retver}"]
    with util.disk_files({"input.json": inp.model_dump_json()}, {}, cwd=tmp_path):
        check_result(run_qcengine_cli(args))

    args = ["run", "psi4", f"--return-version={retver}", "-"]
    # model_dump_json() works on v1 or v2 (see above). below tests that json() still works on v1.
    if "as_v1" in request.node.name:
        check_result(run_qcengine_cli(args, stdin=inp.json()))
    else:
        check_result(run_qcengine_cli(args, stdin=inp.model_dump_json()))


@uusing("geometric")
@uusing("psi4")
def test_run_procedure(tmp_path, schema_versions, request):
    """Tests qcengine run-procedure with geometric, psi4, and JSON input"""
    models, retver, _ = schema_versions

    def check_result(stdout):
        # print(f"STDOUT <<< {stdout} >>>")
        output = json.loads(stdout)
        output = checkver_and_convert(output, request.node.name, "post", cast_dict_as="OptimizationResult")
        assert output["provenance"]["creator"].lower() == "geometric"
        assert output["success"] is True

    if from_v2(request.node.name):
        inp = {
            "specification": {
                "keywords": {"coordsys": "tric", "maxiter": 100},
                "specification": {
                    "driver": "gradient",
                    "model": {"method": "HF", "basis": "sto-3g"},
                    "keywords": {},
                    "program": "psi4",
                },
            },
            "initial_molecule": models.Molecule(**get_molecule("hydrogen", return_dict=True)),
        }
    else:
        inp = {
            "keywords": {"coordsys": "tric", "maxiter": 100, "program": "psi4"},
            "input_specification": {"driver": "gradient", "model": {"method": "HF", "basis": "sto-3g"}, "keywords": {}},
            "initial_molecule": models.Molecule(**get_molecule("hydrogen", return_dict=True)),
        }
    inp = models.OptimizationInput(**inp)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    if "to_v" in request.node.name:
        args = ["run", "geometric", inp.model_dump_json(), f"--return-version={retver}"]
    else:
        args = ["run-procedure", "geometric", inp.model_dump_json()]
    check_result(run_qcengine_cli(args))

    if "to_v" in request.node.name:
        args = ["run", "geometric", f"--return-version={retver}", os.path.join(tmp_path, "input.json")]
    else:
        args = ["run-procedure", "geometric", os.path.join(tmp_path, "input.json")]
    with util.disk_files({"input.json": inp.model_dump_json()}, {}, cwd=tmp_path):
        check_result(run_qcengine_cli(args))

    # try unified route
    args = ["run", "geometric", inp.model_dump_json(), f"--return-version={retver}"]
    check_result(run_qcengine_cli(args, stdin=inp.model_dump_json()))



from typing import Any, ClassVar, Dict, Union

from qcelemental.models.v2 import AtomicInput, AtomicSpecification, OptimizationInput, OptimizationResult, Provenance

from qcengine.config import TaskConfig
from qcengine.exceptions import InputError, UnknownError
from qcengine.procedures.model import ProcedureHarness
from qcengine.procedures.nwchem_opt.harvester import harvest_as_atomic_result
from qcengine.programs.nwchem.runner import NWChemHarness


class NWChemDriverProcedure(ProcedureHarness):
    """Structural relaxation using NWChem's optimizer"""

    _defaults: ClassVar[Dict[str, Any]] = {"name": "NWChemDriver", "procedure": "optimization"}

    def found(self, raise_error: bool = False) -> bool:
        nwc_harness = NWChemHarness()
        return nwc_harness.found(raise_error)

    def get_version(self) -> str:
        nwc_harness = NWChemHarness()
        return nwc_harness.get_version()

    def build_input_model(
        self, data: Union[Dict[str, Any], "OptimizationInput"], *, return_input_schema_version: bool = False
    ) -> "OptimizationInput":
        return self._build_model(data, "OptimizationInput", return_input_schema_version=return_input_schema_version)

    def compute(self, input_model: OptimizationInput, config: TaskConfig) -> "BaseModel":
        nwc_harness = NWChemHarness()
        self.found(raise_error=True)

        # Unify the keywords from the OptimizationInput and QCInputSpecification
        #  Optimization input will override, but don't tell users this as it seems unnecessary
        keywords = input_model.specification.keywords.copy()
        keywords.update(input_model.specification.specification.keywords)
        if (uprog := input_model.specification.specification.program) and (uprog.lower() != "nwchem"):
            raise InputError(f"NWChemDriver procedure only works with NWChem, not {uprog}")

        # Add a flag to the atomic input that tells the NWChemHarness we are calling it from driver
        #  This is needed for the NWCHarness to make some changes to the input file
        input_model.specification.specification.extras["is_driver"] = True

        # Make an atomic input
        atomic_input = AtomicInput(
            molecule=input_model.initial_molecule,
            specification=input_model.specification.specification,
        )

        # Build the inputs for the job
        job_inputs = nwc_harness.build_input(atomic_input, config)

        # As of v2, now use "is_driver" to signal task ... optimize in the first place. This allows driver=gradient to work even though task ... gradient not last in file
        # Replace the last line with a "task {} optimize"
        # input_file: str = job_inputs["infiles"]["nwchem.nw"].strip()
        # beginning, last_line = input_file.rsplit("\n", 1)
        # assert last_line.startswith("task")
        # last_line = f"task {last_line.split(' ')[1]} optimize"
        # job_inputs["infiles"]["nwchem.nw"] = f"{beginning}\n{last_line}"

        # Run it!
        success, dexe = nwc_harness.execute(job_inputs)

        # Check for common errors
        if "There is an error in the input file" in dexe["stdout"]:
            raise InputError(dexe["stdout"])
        if "not compiled" in dexe["stdout"]:
            # recoverable with a different compilation with optional modules
            raise InputError(dexe["stdout"])

        # Parse it
        if success:
            dexe["outfiles"]["stdout"] = dexe["stdout"]
            dexe["outfiles"]["stderr"] = dexe["stderr"]
            return self.parse_output(dexe["outfiles"], input_model)
        else:
            raise UnknownError(dexe["stdout"])

    def parse_output(self, outfiles: Dict[str, str], input_model: OptimizationInput) -> OptimizationResult:
        optsubproperties = ["calcinfo_natom", "return_energy", "return_gradient"]

        # Get the stdout from the calculation (required)
        stdout = outfiles.pop("stdout")
        stderr = outfiles.pop("stderr")

        # Parse out the atomic results from the file
        atomic_results = harvest_as_atomic_result(input_model, stdout)

        # Isolate the converged result
        final_step = atomic_results[-1]

        # input_data = input_model.model_dump()
        # input_data["specification"]["specification"]["program"] = "nwchem"

        return OptimizationResult(
            input_data=input_model,
            final_molecule=final_step.molecule,
            trajectory_results=atomic_results,
            trajectory_properties=[
                {k: v for k, v in grad.properties.model_dump().items() if k in optsubproperties}
                for grad in atomic_results
            ],
            properties={
                "return_energy": atomic_results[-1].properties.return_energy,
                "optimization_iterations": len(atomic_results),
            },
            stdout=stdout,
            stderr=stderr,
            success=True,
            provenance=Provenance(
                creator="NWChemDriver", version=self.get_version(), routine="nwchem_opt"
            ),  # was NWChemRelax which frankly is a better name
        )



import importlib
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Tuple, Union

from qcelemental.util import parse_version, safe_version, which_import

from ..exceptions import InputError
from ..util import model_wrapper
from .model import ProcedureHarness

if TYPE_CHECKING:
    from qcmanybody.models import ManyBodyInput, ManyBodyResult

    from ..config import TaskConfig


class QCManyBodyProcedure(ProcedureHarness):

    _defaults: ClassVar[Dict[str, Any]] = {"name": "QCManyBody", "procedure": "manybody"}

    version_cache: Dict[str, str] = {}

    def found(self, raise_error: bool = False) -> bool:
        return which_import(
            "qcmanybody",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install qcmanybody -c conda-forge`.",
        )

    def build_input_model(
        self, data: Union[Dict[str, Any], "ManyBodyInput"], *, return_input_schema_version: bool = False
    ) -> "ManyBodyInput":

        return self._build_model(data, "ManyBodyInput", return_input_schema_version=return_input_schema_version)

    # temporary can't use procedure/model.py b/c different import location than qcel.models
    def _build_model(
        self, data: Dict[str, Any], model: "BaseModel", /, *, return_input_schema_version: bool = False
    ) -> Union["BaseModel", Tuple["BaseModel", int]]:
        """
        Quick wrapper around util.model_wrapper for inherited classes
        """
        import qcmanybody

        try:
            module_v1 = importlib.import_module("qcmanybody.models.v1")
            module_v2 = importlib.import_module("qcmanybody.models.v2")
            v1_model = getattr(module_v1, model)
            v2_model = getattr(module_v2, model)

        except ModuleNotFoundError:
            module_v1 = importlib.import_module("qcmanybody.models")
            v1_model = getattr(module_v1, model)
            v2_model = None

        qcmb_v1v2 = parse_version(self.get_version()) >= parse_version("0.6a0")

        if isinstance(data, v1_model):
            mdl = model_wrapper(data, v1_model)
        elif v2_model is not None and isinstance(data, v2_model):
            mdl = model_wrapper(data, v2_model)
        elif isinstance(data, dict):
            # remember these are user-provided dictionaries, so they'll have the mandatory fields,
            #   like driver, not the helpful discriminator fields like schema_version.
            # so long as versions distinguishable by a *required* field, id by dict is reliable.

            if data.get("specification", False) or data.get("schema_version") == 2:
                if v2_model is None:
                    raise InputError(
                        "v2 ManyBody inputs require QCManyBody >= 0.6a0 (first version with v2 schemas); "
                        "please upgrade your qcmanybody installation."
                    )
                mdl = model_wrapper(data, v2_model)
            else:
                mdl = model_wrapper(data, v1_model)

        input_schema_version = mdl.schema_version
        if input_schema_version != 1 and qcmb_v1v2 is False:
            raise InputError("Can't use v2 ManyBody")

        if qcmb_v1v2:
            if return_input_schema_version:
                return mdl.convert_v(2), input_schema_version
            else:
                return mdl.convert_v(2)
        else:
            if return_input_schema_version:
                return mdl, input_schema_version
            else:
                return mdl

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("qcmanybody")
        if which_prog not in self.version_cache:
            import qcmanybody

            self.version_cache[which_prog] = safe_version(qcmanybody.__version__)

        return self.version_cache[which_prog]

    def compute(self, input_model: "ManyBodyInput", config: "TaskConfig") -> "ManyBodyResult":
        if input_model.schema_version == 2:
            from qcmanybody.v2 import ManyBodyComputer
        else:
            from qcmanybody import ManyBodyComputer

        output_model = ManyBodyComputer.from_manybodyinput(input_model)

        return output_model
